from __future__ import annotations

import platform
import sys
from time import perf_counter
from typing import TYPE_CHECKING, Optional

from verbatim.logging_utils import get_status_logger
from verbatim_transcript import TranscriberProtocol

from .transcript.sentences import SentenceTokenizer

if TYPE_CHECKING:
    from .voices.silences import VoiceActivityDetection

STATUS_LOG = get_status_logger()


class Models:
    vad: "VoiceActivityDetection"
    sentence_tokenizer: SentenceTokenizer

    def __init__(
        self,
        *,
        device: str,
        whisper_model_size: str = "large-v3",
        stream: bool = False,
        transcriber: Optional[TranscriberProtocol] = None,
        transcriber_backend: str = "auto",
        qwen_asr_model_size: str = "Qwen/Qwen3-ASR-1.7B",
        qwen_aligner_model_size: str = "Qwen/Qwen3-ForcedAligner-0.6B",
        qwen_dtype: str = "auto",
        qwen_max_inference_batch_size: int = 1,
        qwen_max_new_tokens: int = 256,
    ):
        self._transcriber: Optional[TranscriberProtocol] = transcriber
        self._device = device
        self._whisper_model_size = whisper_model_size
        self._config_transcriber_backend = transcriber_backend
        self._qwen_asr_model_size = qwen_asr_model_size
        self._qwen_aligner_model_size = qwen_aligner_model_size
        self._qwen_dtype = qwen_dtype
        self._qwen_max_inference_batch_size = qwen_max_inference_batch_size
        self._qwen_max_new_tokens = qwen_max_new_tokens

        if transcriber is not None:
            STATUS_LOG.info("Using injected transcriber implementation.")
        else:
            STATUS_LOG.info(
                "Transcriber will be lazy-loaded on first use (backend=%s, device=%s, size=%s).",
                transcriber_backend,
                device,
                whisper_model_size,
            )

        STATUS_LOG.info("Lazy-loading Silero VAD model.")
        vad_start = perf_counter()
        from .voices.silences import SileroVoiceActivityDetection  # pylint: disable=import-outside-toplevel

        self.vad: "VoiceActivityDetection" = SileroVoiceActivityDetection()
        STATUS_LOG.info("Silero VAD loaded in %.2fs", perf_counter() - vad_start)

        STATUS_LOG.info("Initializing Sentence Tokenizer.")
        if stream:
            from .transcript.sentences import FastSentenceTokenizer  # pylint: disable=import-outside-toplevel

            self.sentence_tokenizer = FastSentenceTokenizer()
        else:
            STATUS_LOG.info("Lazy-loading SaT sentence tokenizer.")
            sat_start = perf_counter()
            from .transcript.sentences import SaTSentenceTokenizer  # pylint: disable=import-outside-toplevel

            self.sentence_tokenizer = SaTSentenceTokenizer(device)
            STATUS_LOG.info("SaT sentence tokenizer loaded in %.2fs", perf_counter() - sat_start)

    @property
    def transcriber(self) -> TranscriberProtocol:
        if self._transcriber is None:
            transcriber_start = perf_counter()
            self._transcriber = self._build_transcriber()
            STATUS_LOG.info("Transcriber ready in %.2fs", perf_counter() - transcriber_start)
        return self._transcriber

    def _build_transcriber(self) -> TranscriberProtocol:
        # pylint: disable=import-outside-toplevel
        backend = (getattr(self, "_config_transcriber_backend", "auto") or "auto").lower()
        if backend in ("qwen", "qwen-asr"):
            STATUS_LOG.info("Using Qwen3-ASR transcriber.")
            from .voices.transcribe.qwen_asr import QwenAsrTranscriber  # pylint: disable=import-outside-toplevel

            return QwenAsrTranscriber(
                model_size_or_path=self._qwen_asr_model_size,
                aligner_model_size_or_path=self._qwen_aligner_model_size,
                device=self._device,
                dtype=self._qwen_dtype,
                max_inference_batch_size=self._qwen_max_inference_batch_size,
                max_new_tokens=self._qwen_max_new_tokens,
            )

        if sys.platform == "darwin":
            # If this is an Apple Silicon device, use the MLX Whisper transcriber
            if platform.processor() == "arm":
                try:
                    from .voices.transcribe.whispermlx import WhisperMlxTranscriber  # pylint: disable=import-outside-toplevel
                except ImportError:
                    STATUS_LOG.warning("WhisperMLX is not installed; falling back to faster-whisper on Apple Silicon.")
                else:
                    STATUS_LOG.info("Using WhisperMLX transcriber on Apple Silicon")
                    return WhisperMlxTranscriber(model_size_or_path=self._whisper_model_size)
            else:
                raise RuntimeError("Intel macOS is no longer supported; Apple Silicon is required on macOS.")

        STATUS_LOG.info("Using 'faster-whisper' transcriber.")
        from .voices.transcribe.faster_whisper import FasterWhisperTranscriber  # pylint: disable=import-outside-toplevel

        return FasterWhisperTranscriber(model_size_or_path=self._whisper_model_size, device=self._device)



import sys
from typing import List

from .file import FileFormatter, MultiFileFormatter
from .writer import TranscriptWriterConfig


def configure_writers(
    write_config: TranscriptWriterConfig,
    output_formats: List[str],
    original_audio_file: str,
    output_prefix_no_ext: str,
    stdout_enabled: bool = True,
) -> MultiFileFormatter:
    # pylint: disable=import-outside-toplevel
    formatters: List[FileFormatter] = []
    use_stdout = "stdout" in output_formats or "stdout-nocolor" in output_formats
    stdout_nocolor = "stdout-nocolor" in output_formats
    use_stdout = use_stdout and stdout_enabled
    if use_stdout:
        selected_formats = [fmt for fmt in output_formats if fmt not in ("stdout", "stdout-nocolor")]
        if len(selected_formats) != 1:
            raise ValueError("stdout output requires exactly one selected output format")
        stdout_output = sys.stdout.buffer

    if "txt" in output_formats:
        from .stdout import StdoutTranscriptWriter
        from .txt import TextTranscriptWriter

        if use_stdout:
            if stdout_nocolor:
                writer = StdoutTranscriptWriter(config=write_config, with_colours=False)
            else:
                writer = StdoutTranscriptWriter(config=write_config, with_colours=True)
            formatters.append(FileFormatter(writer=writer, output=stdout_output, close_output=False))
        else:
            formatters.append(FileFormatter(writer=TextTranscriptWriter(config=write_config), output_path=f"{output_prefix_no_ext}.txt"))

    if "ass" in output_formats:
        from .ass import AssTranscriptWriter

        if use_stdout:
            formatters.append(
                FileFormatter(
                    writer=AssTranscriptWriter(config=write_config, original_audio_file=original_audio_file),
                    output=stdout_output,
                    close_output=False,
                )
            )
        else:
            formatters.append(
                FileFormatter(
                    writer=AssTranscriptWriter(config=write_config, original_audio_file=original_audio_file),
                    output_path=f"{output_prefix_no_ext}.ass",
                )
            )

    if "docx" in output_formats:
        from .docx import DocxTranscriptWriter

        if use_stdout:
            formatters.append(FileFormatter(writer=DocxTranscriptWriter(config=write_config), output=stdout_output, close_output=False))
        else:
            formatters.append(FileFormatter(writer=DocxTranscriptWriter(config=write_config), output_path=f"{output_prefix_no_ext}.docx"))

    if "md" in output_formats:
        from .md import MarkdownTranscriptWriter

        if use_stdout:
            formatters.append(FileFormatter(writer=MarkdownTranscriptWriter(config=write_config), output=stdout_output, close_output=False))
        else:
            formatters.append(FileFormatter(writer=MarkdownTranscriptWriter(config=write_config), output_path=f"{output_prefix_no_ext}.md"))

    if "json" in output_formats:
        from .json import JsonTranscriptWriter

        if use_stdout:
            formatters.append(FileFormatter(writer=JsonTranscriptWriter(config=write_config), output=stdout_output, close_output=False))
        else:
            formatters.append(FileFormatter(writer=JsonTranscriptWriter(config=write_config), output_path=f"{output_prefix_no_ext}.json"))

    if "jsonl" in output_formats:
        from .json import JsonlTranscriptWriter

        if use_stdout:
            formatters.append(FileFormatter(writer=JsonlTranscriptWriter(config=write_config), output=stdout_output, close_output=False))
        else:
            formatters.append(FileFormatter(writer=JsonlTranscriptWriter(config=write_config), output_path=f"{output_prefix_no_ext}.jsonl"))

    return MultiFileFormatter(formatters=formatters)



import logging
import os
from typing import Optional

import soundfile as sf

from verbatim.cache import ArtifactCache
from verbatim_diarization.diarize.base import DiarizationStrategy
from verbatim_diarization.utils import sanitize_uri_component
from verbatim_files.rttm import Annotation, Segment, dumps_rttm
from verbatim_files.vttm import AudioRef, dumps_vttm

LOG = logging.getLogger(__name__)


class ChannelDiarization(DiarizationStrategy):
    """Treat each channel as a separate speaker without further diarization."""

    def __init__(self, *, cache: ArtifactCache, speaker_labels: Optional[dict[int, str]] = None, speaker: str = "SPEAKER_{idx}", offset: int = 0):
        super().__init__(cache=cache)
        self.speaker_labels = speaker_labels or {}
        self.speaker_pattern = speaker
        self.speaker_offset = offset

    def compute_diarization(
        self,
        file_path: str,
        out_rttm_file: Optional[str] = None,
        out_vttm_file: Optional[str] = None,
        status_hook=None,
        **kwargs,
    ) -> Annotation:
        _ = status_hook, kwargs
        buffer = self.cache.bytes_io(file_path)
        audio, sample_rate = sf.read(buffer)
        buffer.seek(0)
        audio_info = sf.info(buffer)
        LOG.info("Input file channels: %s, sample rate: %s", audio_info.channels, audio_info.samplerate)

        if audio.ndim == 1:
            # Expand mono to (samples, 1) so channel labeling still applies
            audio = audio[:, None]
        if audio.ndim < 2 or audio.shape[1] < 1:
            raise ValueError("Channel diarization requires audio with at least one channel")

        uri = sanitize_uri_component(os.path.splitext(os.path.basename(file_path))[0])
        duration = len(audio) / sample_rate

        segments = []
        audio_refs = []
        num_channels = audio.shape[1]
        for idx in range(num_channels):
            speaker = self.speaker_labels.get(idx, self.speaker_pattern.format(idx=idx + self.speaker_offset))
            segments.append(Segment(start=0.0, end=duration, speaker=speaker, file_id=uri))
            channel_id = sanitize_uri_component(f"{uri}_ch{idx}", fallback=f"{uri}_{idx}")
            audio_refs.append(AudioRef(id=channel_id, path=file_path, channels=str(idx)))

        annotation = Annotation(segments=segments, file_id=uri)

        if out_rttm_file:
            self.cache.set_text(out_rttm_file, dumps_rttm(annotation))
            LOG.info("Wrote channel diarization to RTTM file: %s", out_rttm_file)

        if out_vttm_file:
            self.cache.set_text(out_vttm_file, dumps_vttm(audio=audio_refs, annotation=annotation))
            LOG.info("Wrote channel diarization to VTTM file: %s", out_vttm_file)

        return annotation



"""Parse diarization policy strings into strategy assignments."""

from dataclasses import dataclass
from typing import Dict, List, Optional, Set


@dataclass
class PolicyClause:
    targets: Set[int]  # empty set means wildcard
    strategy: str
    params: Dict[str, str]


def parse_targets(target_str: str) -> Set[int]:
    target_str = target_str.strip()
    if target_str == "*":
        return set()
    targets: Set[int] = set()
    for part in target_str.split(","):
        part = part.strip()
        if not part:
            continue
        if "-" in part:
            start_s, end_s = part.split("-", 1)
            start = int(start_s)
            end = int(end_s)
            if end < start:
                start, end = end, start
            targets.update(range(start, end + 1))
        else:
            targets.add(int(part))
    return targets


def parse_params(param_str: str) -> Dict[str, str]:
    params: Dict[str, str] = {}
    for pair in param_str.split("&"):
        if not pair:
            continue
        if "=" in pair:
            k, v = pair.split("=", 1)
            params[k.strip()] = v.strip()
        else:
            params[pair.strip()] = ""
    return params


def parse_clause(clause: str) -> PolicyClause:
    clause = clause.strip()
    target_part, rest = clause.split("=", 1) if "=" in clause else ("*", clause)
    if "?" in rest:
        strategy_part, param_part = rest.split("?", 1)
        params = parse_params(param_part)
    else:
        strategy_part = rest
        params = {}
    targets = parse_targets(target_part)
    strategy = strategy_part.strip()
    return PolicyClause(targets=targets, strategy=strategy, params=params)


def parse_policy(policy: str) -> List[PolicyClause]:
    if not policy:
        return []
    clauses = []
    for raw in policy.split(";"):
        raw = raw.strip()
        if not raw:
            continue
        clauses.append(parse_clause(raw))
    return clauses


def assign_channels(clauses: List[PolicyClause], nchannels: int) -> Dict[int, PolicyClause]:
    assignments: Dict[int, PolicyClause] = {}
    wildcard: Optional[PolicyClause] = None
    for clause in clauses:
        if len(clause.targets) == 0:
            wildcard = clause
            continue
        for ch in clause.targets:
            assignments[ch] = clause
    if wildcard:
        for ch in range(nchannels):
            if ch not in assignments:
                assignments[ch] = wildcard
    return assignments



import logging
import sys

LOG = logging.getLogger(__name__)


def main():
    # pylint: disable=import-outside-toplevel
    from argparse import Namespace

    from verbatim.logging_utils import configure_status_logger, get_status_logger
    from verbatim_cli.args import build_parser
    from verbatim_cli.config_file import load_config_file, merge_args, select_profile
    from verbatim_cli.configure import (
        compute_log_level,
        make_config,
        resolve_speakers,
    )
    from verbatim_cli.env import load_env_file
    from verbatim_cli.run_single import run_single_input

    parser = build_parser(prog="verbatim")
    base_defaults: Namespace = parser.parse_args([])
    user_args = parser.parse_args()

    cfg_data = {}
    if getattr(user_args, "config", None):
        cfg_data = load_config_file(user_args.config)

    profile_overrides = select_profile(cfg_data, filename=user_args.input)
    args = merge_args(base_defaults, profile_overrides, user_args)

    log_level = compute_log_level(args.verbose)
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)

    log_format = "%(asctime)s [%(levelname)s][%(filename)s:%(lineno)d][%(funcName)s] %(message)s"
    formatter = logging.Formatter(fmt=log_format, datefmt="%Y-%m-%dT%H:%M:%SZ")
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(log_level)
    stderr_handler.setFormatter(formatter)
    handlers = [stderr_handler]
    status_file_handler = None
    if getattr(args, "log_file", None):
        file_handler = logging.FileHandler(args.log_file, mode="w", encoding="utf-8")
        file_handler.setLevel(log_level)
        file_handler.setFormatter(formatter)
        handlers.append(file_handler)

        status_file_handler = logging.FileHandler(args.log_file, mode="a", encoding="utf-8")
        status_file_handler.setLevel(logging.DEBUG if args.verbose >= 2 else logging.INFO)
        status_file_handler.setFormatter(formatter)

    logging.basicConfig(level=log_level, handlers=handlers, force=True)
    configure_status_logger(verbose=args.verbose, fmt=log_format, datefmt="%Y-%m-%dT%H:%M:%SZ", file_handler=status_file_handler)
    status_log = get_status_logger()

    # load the values from the .env file, if present
    load_env_file()

    # Check if the version option is specified
    if hasattr(args, "version") and args.version:
        return  # Exit if the version option is specified

    config = make_config(args)

    if args.input is None and not args.install:
        parser.print_usage(sys.stderr)
        LOG.error("Missing input path. Provide an audio file, '-' for stdin, or '>' for microphone.")
        return

    # Handle install-only mode
    if args.install:
        status_log.info("Installing/prefetching models into cache...")
        from verbatim.prefetch import prefetch

        prefetch(model_cache_dir=args.model_cache, whisper_size=config.whisper_model_size)
        status_log.info("Model prefetch complete.")
        return

    source_path = args.input
    speakers = resolve_speakers(args)

    status_log.info(
        "Diarization settings: strategy=%s speakers=%s vttm=%s rttm=%s",
        args.diarize,
        speakers,
        args.vttm,
        args.diarization,
    )

    run_single_input(
        args=args,
        log_level=log_level,
        source_path=source_path,
        config=config,
        default_stdout=True,
    )


if __name__ == "__main__":
    main()



"""Helpers for normalizing pyannote pipeline outputs."""

from typing import Any


def select_speaker_diarization(output: Any) -> Any:
    """Prefer exclusive diarization when pyannote exposes it."""
    if hasattr(output, "exclusive_speaker_diarization"):
        exclusive = output.exclusive_speaker_diarization  # type: ignore[attr-defined]
        if exclusive is not None:
            return exclusive

    if hasattr(output, "speaker_diarization"):
        diarization = output.speaker_diarization  # type: ignore[attr-defined]
        if diarization is not None:
            return diarization

    return output



from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
)

from numpy.typing import NDArray

# Minimal contract for transcription backends; lets us swap local/remote implementations.
if TYPE_CHECKING:
    from verbatim.transcript.words import Word
else:  # pragma: no cover - type-only fallback to avoid heavy imports
    Word = object  # pylint: disable=invalid-name


class TranscriberProtocol(Protocol):
    def guess_language(self, audio: NDArray, lang: List[str]) -> Tuple[str, float]: ...

    def transcribe(
        self,
        *,
        audio: NDArray,
        lang: str,
        prompt: str,
        prefix: str,
        window_ts: int,
        audio_ts: int,
        whisper_beam_size: int = 3,
        whisper_best_of: int = 3,
        whisper_patience: float = 1.0,
        whisper_temperatures: Optional[List[float]] = None,
    ) -> List[Word]: ...


class VadProtocol(Protocol):
    def __call__(self, audio: NDArray, min_speech_duration_ms: int, min_silence_duration_ms: int) -> List[Dict[str, int]]: ...


# Callbacks are intentionally simple so they can be implemented without heavy deps.
VadFn = VadProtocol
GuessLanguageFn = Callable[[NDArray, List[str]], Tuple[str, float]]

TUtt_co = TypeVar("TUtt_co", covariant=True)  # pylint: disable=invalid-name
TWord_co = TypeVar("TWord_co", covariant=True)  # pylint: disable=invalid-name


@dataclass
class LanguageDetectionRequest:
    audio: NDArray
    lang: List[str]
    timestamp: int
    window_ts: int
    audio_ts: int


@dataclass
class LanguageDetectionResult:
    language: str
    probability: float
    samples_used: int


@dataclass
class TranscriptionWindowResult(Generic[TUtt_co, TWord_co]):
    utterance: TUtt_co
    unacknowledged: Sequence[TUtt_co]
    unconfirmed_words: Sequence[TWord_co]

    def as_tuple(self):
        return self.utterance, list(self.unacknowledged), list(self.unconfirmed_words)



import logging
from abc import ABC, abstractmethod
from typing import List, Optional

from verbatim.cache import ArtifactCache
from verbatim_audio.sources.audiosource import AudioSource

# Configure logger
LOG = logging.getLogger(__name__)


class SeparationStrategy(ABC):
    def __init__(self, *, cache: ArtifactCache):
        self.cache = cache

    def __enter__(self) -> "SeparationStrategy":
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        return False

    @abstractmethod
    def separate_speakers(
        self,
        *,
        file_path: str,
        out_rttm_file: Optional[str] = None,
        out_vttm_file: Optional[str] = None,
        out_speaker_wav_prefix="",
        nb_speakers: Optional[int] = None,
        start_sample: int = 0,
        end_sample: Optional[int] = None,
    ) -> List[AudioSource]:
        pass



from abc import ABC, abstractmethod


class IdProvider(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def next(self) -> str:
        pass


class CounterIdProvider(IdProvider):
    prefix: str
    suffix: str
    counter: int

    def __init__(self, prefix: str = "", suffix: str = ""):
        self.prefix = prefix
        self.suffix = suffix
        self.counter = 0

    def next(self) -> str:
        self.counter += 1
        return f"{self.prefix}{self.counter}{self.suffix}"



import dataclasses
import logging
from dataclasses import dataclass, field
from typing import Any, List

import tqdm

from ..transcript.words import Utterance
from .diarization_metrics import UtteranceMetrics, compute_utterance_metrics


@dataclass
class UtteranceMetricsWithId(UtteranceMetrics):
    utterance_id: str = ""


@dataclass
class Metrics:
    # pylint: disable=invalid-name
    utterances: List[UtteranceMetricsWithId] = field(default_factory=list)
    WER: float = 0.0  # Word Error Rate - Measures transcription accuracy
    WDER: float = 0.0  # Word Diarization Error Rate - Measures speaker attribution errors
    cpWER: float = 0.0  # Concatenated-permutation WER - An enhanced WER that accounts for speaker permutation errors
    SpkCntMAE: float = 0.0  # Mean Absolute Error - Measures differences in estimated and actual speaker counts


def compute_metrics_summary(
    json_dict: dict[str, Any],
    ref_text_field: str = "ref_text",
    hyp_text_field: str = "hyp_text",
    ref_spk_field: str = "ref_spk",
    hyp_spk_field: str = "hyp_spk",
) -> Metrics:
    """Compute metrics for all utterances in a json object."""
    compute_diarization_metrics = bool(ref_spk_field) and bool(hyp_spk_field)
    if compute_diarization_metrics:
        for utt in json_dict.get("utterances", []):
            hyp_spk = utt.get(hyp_spk_field)
            ref_spk = utt.get(ref_spk_field)
            if not hyp_spk and not ref_spk:
                compute_diarization_metrics = False
                break
            if not (hyp_spk and ref_spk):
                LOG.warning("Speaker labels missing on one side; skipping diarization metrics for this comparison.")
                compute_diarization_metrics = False
                break
    result_dict = Metrics()
    for utt in tqdm.tqdm(json_dict["utterances"]):
        if compute_diarization_metrics:
            utt_metrics = compute_utterance_metrics(
                hyp_text=utt[hyp_text_field],
                ref_text=utt[ref_text_field],
                hyp_spk=utt[hyp_spk_field],
                ref_spk=utt[ref_spk_field],
            )
        else:
            utt_metrics = compute_utterance_metrics(
                hyp_text=utt[hyp_text_field],
                ref_text=utt[ref_text_field],
            )
        utt_result = UtteranceMetricsWithId(**dataclasses.asdict(utt_metrics))
        utt_result.utterance_id = utt["utterance_id"]
        result_dict.utterances.append(utt_result)

    final_wer_total = 0
    final_wer_correct = 0
    final_wer_sub = 0
    final_wer_delete = 0
    final_wer_insert = 0
    final_wder_total = 0
    final_wder_correct = 0
    final_wder_sub = 0
    final_cpwer_total = 0
    final_cpwer_correct = 0
    final_cpwer_sub = 0
    final_cpwer_delete = 0
    final_cpwer_insert = 0
    final_speaker_count_absolute_error_total = 0
    for utt in result_dict.utterances:
        final_wer_total += utt.wer_total
        final_wer_correct += utt.wer_correct
        final_wer_sub += utt.wer_sub
        final_wer_delete += utt.wer_delete
        final_wer_insert += utt.wer_insert
        if compute_diarization_metrics:
            final_wder_total += utt.wder_total
            final_wder_correct += utt.wder_correct
            final_wder_sub += utt.wder_sub
            final_cpwer_total += utt.cpwer_total
            final_cpwer_correct += utt.cpwer_correct
            final_cpwer_sub += utt.cpwer_sub
            final_cpwer_delete += utt.cpwer_delete
            final_cpwer_insert += utt.cpwer_insert
            final_speaker_count_absolute_error_total += abs(utt.speaker_count_error)

    result_dict.WER = (final_wer_sub + final_wer_delete + final_wer_insert) / final_wer_total

    if compute_diarization_metrics:
        result_dict.WDER = final_wder_sub / final_wder_total
        result_dict.cpWER = (final_cpwer_sub + final_cpwer_delete + final_cpwer_insert) / final_cpwer_total
        result_dict.SpkCntMAE = final_speaker_count_absolute_error_total / len(result_dict.utterances)
    return result_dict


LOG = logging.getLogger(__name__)


def compute_metrics(hyp_data: List[Utterance], ref_data: List[Utterance]) -> Metrics:
    """Calculate metrics between hypothesis and reference data."""
    data = {}

    def get_speaker_id(name: str, cache: List[str]):
        for idx, speaker in enumerate(cache):
            if speaker == name:
                return idx + 1  # Return a 1-indexed ID if the name exists.

        cache.append(name)
        return len(cache)  # The new speaker's ID is the new length of the cache.

    speakers = []
    ref_speakers = []
    data["utterances"] = [
        {
            "utterance_id": "utt0",
            "hyp_text": " ".join([utt.text for utt in hyp_data]),
            "hyp_spk": " ".join([f"{get_speaker_id(utt.speaker or 'None', speakers)} " * len(utt.text.split()) for utt in hyp_data]),
            "ref_text": " ".join([utt.text for utt in ref_data]),
            "ref_spk": " ".join([f"{get_speaker_id(utt.speaker or 'None', ref_speakers)} " * len(utt.text.split()) for utt in ref_data]),
        }
    ]

    metrics: Metrics = compute_metrics_summary(data)
    return metrics



from typing import List, Optional

from verbatim.transcript.words import Utterance, Word

from .writer import TranscriptWriter, TranscriptWriterConfig


class MultiTranscriptWriter(TranscriptWriter):
    def __init__(self, writers: Optional[List[TranscriptWriter]] = None):
        super().__init__(config=TranscriptWriterConfig())
        self.writers = []
        if writers is not None:
            self.writers += writers

    def get_extension(self):
        return ""

    def format_utterance(
        self,
        utterance: Utterance,
        unacknowledged_utterance: Optional[List[Utterance]] = None,
        unconfirmed_words: Optional[List[Word]] = None,
    ) -> bytes:
        for w in self.writers:
            w.format_utterance(
                utterance=utterance,
                unacknowledged_utterance=unacknowledged_utterance,
                unconfirmed_words=unconfirmed_words,
            )
        return b""

    def flush(self) -> bytes:
        for w in self.writers:
            w.flush()
        return b""

    def add_writer(self, writer: TranscriptWriter):
        self.writers.append(writer)



# pylint: disable=too-many-lines
import logging
import math
import os
import sys
import traceback
import wave
from dataclasses import dataclass, field
from io import StringIO
from time import perf_counter
from typing import Any, Generator, List, Optional, TextIO, Tuple

import numpy as np
from colorama import Fore
from numpy.typing import NDArray

from verbatim.logging_utils import get_status_logger
from verbatim.non_speech import create_non_speech_classifier
from verbatim_audio.audio import samples_to_seconds
from verbatim_audio.sources.audiosource import AudioSource, AudioStream
from verbatim_files.format.factory import configure_writers
from verbatim_files.format.json import read_utterances
from verbatim_files.format.txt import (
    COLORSCHEME_ACKNOWLEDGED,
    COLORSCHEME_NONE,
    COLORSCHEME_UNACKNOWLEDGED,
    COLORSCHEME_UNCONFIRMED,
    TranscriptFormatter,
)
from verbatim_files.format.writer import (
    LanguageStyle,
    ProbabilityStyle,
    SpeakerStyle,
    TimestampStyle,
    TranscriptWriterConfig,
)
from verbatim_transcript import (
    LanguageDetectionRequest,
    LanguageDetectionResult,
    TranscriberProtocol,
    TranscriptionWindowResult,
    VadFn,
    detect_language,
)

from .config import Config
from .language_id import create_language_identifier
from .models import Models
from .status_types import StatusHook, StatusProgress, StatusUpdate
from .transcript.idprovider import CounterIdProvider, IdProvider
from .transcript.sentences import BoundedSentenceTokenizer, SentenceTokenizer
from .transcript.words import Utterance, Word

# pylint: disable=unused-import
from .voices.transcribe.transcribe import APPEND_PUNCTUATIONS, PREPEND_PUNCTUATIONS

# Optional type-only import to avoid pyannote dependency at runtime
Annotation = Any  # pylint: disable=invalid-name

# Configure logger
LOG = logging.getLogger(__name__)
STATUS_LOG = get_status_logger()
SKIP_MARKER_MIN_DURATION_SAMPLES = int(2.5 * 16000)
SKIP_NOISE_RMS_THRESHOLD = 0.01
_NON_SPEECH_CLASSIFIER_FAILED = object()


@dataclass
class WhisperHistory:
    size: int = 12
    transcript_history: List[List[Word]] = field(default_factory=list)

    @staticmethod
    def _words_summary(words: List[Word], max_text: int = 60) -> str:
        if len(words) == 0:
            return "[]"
        text = "".join(w.word for w in words).replace(os.linesep, " ")
        if len(text) > max_text:
            text = text[: max_text - 3] + "..."
        return f"[{words[0].start_ts}-{words[-1].end_ts}] n={len(words)} '{text}'"

    @staticmethod
    def _words_text(words: List[Word]) -> str:
        return "".join(w.word for w in words).replace(os.linesep, " ")

    def add(self, transcription: List[Word]):
        self.transcript_history.append(transcription)
        if len(self.transcript_history) > self.size:
            self.transcript_history.pop(0)

    @staticmethod
    def advance_transcript(timestamp: int, transcript: List[Word]) -> List[Word]:
        return [w for w in transcript if w.end_ts >= timestamp]

    def advance(self, timestamp: int):
        previous_history = self.transcript_history
        advanced_history = [self.advance_transcript(timestamp, transcript) for transcript in previous_history]
        self.transcript_history = [h for h in advanced_history if len(h) > 0]
        if LOG.isEnabledFor(logging.DEBUG):
            changes = []
            for before, after in zip(previous_history, advanced_history):
                if before != after:
                    changes.append(f"{self._words_summary(before)} -> {self._words_summary(after)}")
            if changes:
                LOG.debug("Advanced transcript history at %d: %s", timestamp, " | ".join(changes))

    @staticmethod
    def confirm_transcript(
        current_words: List[Word],
        transcript: List[Word],
        prefix: List[Word],
        after_ts: int,
    ) -> List[Word]:
        c_index = 0
        while c_index < len(current_words):
            if (c_index < len(prefix) and prefix[c_index].word == current_words[c_index].word) or (current_words[c_index].start_ts <= after_ts):
                c_index = c_index + 1
            else:
                break

        p_index = 0
        while p_index < len(transcript):
            if (p_index < len(prefix) and prefix[p_index].word == transcript[p_index].word) or (transcript[p_index].start_ts <= after_ts):
                p_index = p_index + 1
            else:
                break

        confirmed_words = []

        while p_index < len(transcript) and c_index < len(current_words):
            if transcript[p_index].word.strip().lower() == current_words[c_index].word.strip().lower():
                confirmed_words += [current_words[c_index]]
                p_index += 1
                c_index += 1
            else:
                break
        if len(confirmed_words) > 0:
            LOG.debug(
                "Confirmed candidate: matched=%d transcript=%s current=%s confirmed=%s",
                len(confirmed_words),
                WhisperHistory._words_summary(transcript),
                WhisperHistory._words_summary(current_words),
                WhisperHistory._words_summary(confirmed_words),
            )
        return confirmed_words

    def confirm(
        self,
        current_words: List[Word],
        prefix: List[Word],
        after_ts: int,
    ) -> List[Word]:
        return max(
            (
                self.confirm_transcript(
                    current_words=current_words,
                    transcript=transcript,
                    prefix=prefix,
                    after_ts=after_ts,
                )
                for transcript in self.transcript_history
            ),
            key=len,
            default=[],
        )


@dataclass
class SkipResult:
    next_ts: int
    marker: Optional[Utterance] = None


class RollingWindow:
    array: NDArray

    def __init__(self, window_size, dtype=np.float32):
        self.reset(window_size=window_size, dtype=dtype)

    def reset(self, window_size=-1, dtype=None):
        if window_size == -1:
            window_size = len(self.array)
        if dtype is None:
            dtype = self.array.dtype
        self.array = np.zeros(window_size, dtype=dtype)


@dataclass
class State:
    confirmed_ts: int
    acknowledged_ts: int
    window_ts: int
    audio_ts: int
    processing_started_at: float
    processing_started_window_ts: int
    transcript_candidate_history: WhisperHistory
    rolling_window: RollingWindow
    acknowledged_words: List[Word]
    unconfirmed_words: List[Word]
    acknowledged_utterances: List[Utterance]
    unacknowledged_utterances: List[Utterance]
    last_transcribe_lang: str
    skip_silences: bool = True
    speaker_embeddings: Optional[List] = None
    working_prefix_no_ext: str = "out"
    timing_totals_ms: dict[str, float] = field(init=False)
    last_transcribe_metrics_ms: dict[str, float] = field(init=False)

    def __init__(self, config: Config, working_prefix_no_ext: str = "out"):
        self.config = config
        self.confirmed_ts = -1
        self.acknowledged_ts = -1
        self.window_ts = 0  # Window timestamp in samples
        self.audio_ts = 0  # Global timestamp in samples
        self.processing_started_at = perf_counter()
        self.processing_started_window_ts = 0
        self.transcript_candidate_history = WhisperHistory()
        self.acknowledged_words = []
        self.unconfirmed_words = []
        self.acknowledged_utterances = []
        self.unacknowledged_utterances = []
        self.last_transcribe_lang = ""
        self.speaker_embeddings = []
        self.working_prefix_no_ext = working_prefix_no_ext
        self.utterance_id: IdProvider = CounterIdProvider(prefix="utt")
        self.timing_totals_ms = {
            "detect": 0.0,
            "transcribe": 0.0,
            "confirm": 0.0,
            "sentence": 0.0,
            "acknowledge": 0.0,
            "speaker": 0.0,
            "output": 0.0,
            "advance": 0.0,
            "total": 0.0,
        }
        self.last_transcribe_metrics_ms = {
            "detect_ms": 0.0,
            "transcribe_ms": 0.0,
            "confirm_ms": 0.0,
        }

        window_size = self.config.sampling_rate * self.config.window_duration  # Total samples in 30 seconds
        self.rolling_window = RollingWindow(window_size=window_size, dtype=np.float32)  # Initialize empty rolling window

    def advance_audio_window(self, offset: int):
        if offset <= 0:
            return
        # Do not advance beyond available audio; keep window_ts <= audio_ts
        max_advance = max(0, self.audio_ts - self.window_ts)
        if offset > max_advance:
            LOG.warning(
                f"advance_audio_window requested {samples_to_seconds(offset)}s but only {samples_to_seconds(max_advance)}s available; clamping."
            )
            offset = max_advance
        if offset == 0:
            return
        LOG.debug(
            f"Shifting rolling window by {offset} samples ({samples_to_seconds(offset)}s) "
            f"to offset {self.window_ts + offset} ({samples_to_seconds(self.window_ts + offset)}s); "
            f"Window now has {samples_to_seconds(self.audio_ts - self.window_ts)} seconds of audio."
        )

        self.rolling_window.array = np.roll(self.rolling_window.array, -offset)
        self.rolling_window.array[-offset:] = 0
        self.window_ts += offset

        LOG.debug("Window now has %s seconds of audio.", samples_to_seconds(self.audio_ts - self.window_ts))

    def append_audio_to_window(self, audio_chunk: NDArray):
        if audio_chunk.size == 0:
            LOG.warning("Received empty audio chunk, skipping.")
            return

        # Convert stereo to mono if necessary
        LOG.debug(f"Audio chunk shape before mono conversion: {audio_chunk.shape}")
        if len(audio_chunk.shape) > 1 and audio_chunk.shape[1] > 1:
            audio_chunk = np.mean(audio_chunk, axis=1)
        LOG.debug(f"Audio chunk shape after mono conversion: {audio_chunk.shape}")

        chunk_size = len(audio_chunk)
        window_size = len(self.rolling_window.array)
        # Guard against a temporarily inconsistent state where window_ts > audio_ts
        insertion_start = max(0, self.audio_ts - self.window_ts)
        insertion_end = insertion_start + chunk_size
        if self.audio_ts + chunk_size <= self.window_ts + window_size:
            # Append to the current rolling window
            self.rolling_window.array[insertion_start:insertion_end] = audio_chunk
        else:
            # Shift the window and append new audio
            shift_amount = self.audio_ts + chunk_size - (self.window_ts + window_size)
            LOG.warning(f"failed to acknowledge transcription within audio buffer of {samples_to_seconds(window_size)} seconds")
            self.advance_audio_window(shift_amount)
            self.skip_silences = True
            self.rolling_window.array[-chunk_size:] = audio_chunk

        self.audio_ts += chunk_size


class Verbatim:
    state: State
    config: Config
    vad_callback: Optional[VadFn]
    status_hook: Optional[StatusHook]

    def __init__(
        self,
        config: Config,
        *,
        models=None,
        vad_callback: Optional[VadFn] = None,
        transcriber: Optional[TranscriberProtocol] = None,
        status_hook: Optional[StatusHook] = None,
    ):
        self.config = config
        self.state = State(config)
        self.status_hook = status_hook
        self._non_speech_classifier = None
        if models is None:
            models = Models(
                device=config.device,
                whisper_model_size=config.whisper_model_size,
                stream=config.stream,
                transcriber=transcriber,
                transcriber_backend=config.transcriber_backend,
                qwen_asr_model_size=config.qwen_asr_model_size,
                qwen_aligner_model_size=config.qwen_aligner_model_size,
                qwen_dtype=config.qwen_dtype,
                qwen_max_inference_batch_size=config.qwen_max_inference_batch_size,
                qwen_max_new_tokens=config.qwen_max_new_tokens,
            )
        self.models = models
        self.language_identifier = create_language_identifier(config=config, models=self.models)
        if vad_callback is not None:
            self.vad_callback = vad_callback
        elif hasattr(self.models, "vad"):
            # default Silero VAD path; keeps current behavior while enabling injection later
            self.vad_callback = self.models.vad.find_activity  # type: ignore[attr-defined]
        else:
            self.vad_callback = None

    def _emit_status(self, update: StatusUpdate) -> None:
        if self.status_hook is None:
            return
        try:
            self.status_hook(update)
        except Exception:  # pylint: disable=broad-exception-caught
            LOG.exception("Status hook failed")

    def _emit_utterance_status(
        self,
        *,
        utterance: Utterance,
        unacknowledged: List[Utterance],
        unconfirmed: List[Word],
    ) -> None:
        self._emit_status(
            StatusUpdate(
                state="transcribing",
                utterance=utterance,
                unacknowledged_utterances=unacknowledged,
                unconfirmed_words=unconfirmed,
            )
        )

    def _get_non_speech_classifier(self):
        if self._non_speech_classifier is _NON_SPEECH_CLASSIFIER_FAILED:
            return None
        if self._non_speech_classifier is None:
            try:
                self._non_speech_classifier = create_non_speech_classifier(
                    backend=self.config.non_speech_backend,
                    device=self.config.device,
                    model_name=self.config.ast_audio_model_size,
                )
            except Exception:  # pylint: disable=broad-exception-caught
                self._non_speech_classifier = _NON_SPEECH_CLASSIFIER_FAILED
                raise
        return self._non_speech_classifier

    @staticmethod
    def _get_total_samples(audio_stream: AudioStream) -> Optional[int]:
        total_samples = getattr(audio_stream, "total_samples", None)
        if total_samples is None:
            return None
        start_offset = getattr(audio_stream, "start_offset", 0) or 0
        end_sample = getattr(audio_stream, "end_sample", None)
        if end_sample is not None:
            total_samples = min(total_samples, end_sample)
        return max(0, total_samples - start_offset)

    def _emit_transcribing_progress(self, audio_stream: AudioStream) -> None:
        total_samples = self._get_total_samples(audio_stream)
        if total_samples is None or total_samples <= 0:
            return
        start_offset = getattr(audio_stream, "start_offset", 0) or 0
        current_samples = max(0, self.state.window_ts - start_offset)
        self._emit_status(
            StatusUpdate(
                state="transcribing",
                progress=StatusProgress(
                    current=samples_to_seconds(current_samples),
                    finish=samples_to_seconds(total_samples),
                    units="seconds",
                ),
            )
        )

    def _detect_language_for_audio(self, *, audio: NDArray, window_ts: int, audio_ts: int, timestamp: int) -> LanguageDetectionResult:
        request = LanguageDetectionRequest(
            audio=audio,
            lang=self.config.lang,
            timestamp=timestamp,
            window_ts=window_ts,
            audio_ts=audio_ts,
        )
        return detect_language(request=request, guess_fn=self.language_identifier.guess_language)

    def _read_full_audio(self, audio_stream: AudioStream) -> NDArray:
        chunks: List[NDArray] = []
        chunk_length = max(1, self.config.window_duration)
        while audio_stream.has_more():
            chunk = audio_stream.next_chunk(chunk_length=chunk_length)
            if chunk.size == 0:
                continue
            if len(chunk.shape) > 1 and chunk.shape[1] > 1:
                chunk = np.mean(chunk, axis=1)
            chunks.append(chunk.astype(np.float32))
        if len(chunks) == 0:
            return np.zeros(0, dtype=np.float32)
        return np.concatenate(chunks, axis=0)

    @staticmethod
    def _summarize_words(words: List[Word], max_text: int = 80) -> str:
        if len(words) == 0:
            return "[]"
        text = "".join(word.word for word in words).replace(os.linesep, " ")
        if len(text) > max_text:
            text = text[: max_text - 3] + "..."
        langs = sorted({word.lang for word in words if word.lang})
        langs_summary = f" langs={','.join(langs)}" if langs else ""
        return f"[{words[0].start_ts}-{words[-1].end_ts}] n={len(words)}{langs_summary} '{text}'"

    @staticmethod
    def _words_text(words: List[Word]) -> str:
        return "".join(word.word for word in words).replace(os.linesep, " ")

    @staticmethod
    def _summarize_utterance(utterance: Utterance, max_text: int = 80) -> str:
        text = utterance.text.replace(os.linesep, " ")
        if len(text) > max_text:
            text = text[: max_text - 3] + "..."
        return f"{utterance.utterance_id}[{utterance.start_ts}-{utterance.end_ts}] '{text}'"

    @classmethod
    def _summarize_utterances(cls, utterances: List[Utterance], max_items: int = 4) -> str:
        if len(utterances) == 0:
            return "[]"
        parts = [cls._summarize_utterance(utterance) for utterance in utterances[:max_items]]
        if len(utterances) > max_items:
            parts.append(f"...(+{len(utterances) - max_items})")
        return "[" + ", ".join(parts) + "]"

    @staticmethod
    def _energy_skip_label(samples: NDArray[np.float32]) -> str:
        if len(samples) == 0:
            rms = 0.0
        else:
            rms = float(np.sqrt(np.mean(np.square(samples.astype(np.float64)))))
        return "[ENVIRONMENT NOISE]" if rms >= SKIP_NOISE_RMS_THRESHOLD else "[SILENCE]"

    @staticmethod
    def _format_non_speech_labels(labels: List[str]) -> str:
        if any(label == "silence" for label in labels):
            return "[SILENCE]"
        normalized = []
        for label in labels:
            if not label or label == "other":
                continue
            rendered = label.replace("_", " ").upper()
            if rendered not in normalized:
                normalized.append(rendered)
        if not normalized:
            return "[ENVIRONMENT NOISE]"
        return "[" + ", ".join(normalized) + "]"

    def _build_skip_marker(self, start_ts: int, end_ts: int, samples: NDArray[np.float32]) -> Optional[Utterance]:
        skipped_samples = max(0, end_ts - start_ts)
        if skipped_samples < SKIP_MARKER_MIN_DURATION_SAMPLES:
            LOG.debug(
                "Skipped non-speech span below marker threshold: start=%.2fs end=%.2fs duration=%.2fs threshold=%.2fs",
                samples_to_seconds(start_ts),
                samples_to_seconds(end_ts),
                samples_to_seconds(skipped_samples),
                samples_to_seconds(SKIP_MARKER_MIN_DURATION_SAMPLES),
            )
            return None

        label = self._energy_skip_label(samples)
        if self.config.non_speech_backend == "ast" and label != "[SILENCE]":
            try:
                classifier = self._get_non_speech_classifier()
                if classifier is not None:
                    labels = classifier.classify(samples, self.config.sampling_rate)
                    label = self._format_non_speech_labels(labels)
            except Exception:  # pylint: disable=broad-exception-caught
                self._non_speech_classifier = _NON_SPEECH_CLASSIFIER_FAILED
                LOG.exception("AST non-speech classification failed; falling back to energy-based label")
        elif self.config.non_speech_backend == "ast":
            LOG.debug(
                "Skipping AST non-speech classification for silence-like span: start=%.2fs end=%.2fs",
                samples_to_seconds(start_ts),
                samples_to_seconds(end_ts),
            )
        rms = 0.0 if len(samples) == 0 else float(np.sqrt(np.mean(np.square(samples.astype(np.float64)))))
        LOG.info(
            "Long non-speech skip detected: kind=%s start=%.2fs end=%.2fs duration=%.2fs rms=%.5f",
            label[1:-1].lower(),
            samples_to_seconds(start_ts),
            samples_to_seconds(end_ts),
            samples_to_seconds(skipped_samples),
            rms,
        )
        return Utterance.marker(
            utterance_id=self.state.utterance_id.next(),
            start_ts=start_ts,
            end_ts=end_ts,
            text=label,
            speaker=None,
        )

    def skip_leading_silence(self, max_skip: int, min_speech_duration_ms: int = 500) -> SkipResult:
        if self.vad_callback is None:
            LOG.debug("VAD callback not configured; skipping silence detection and keeping current window.")
            return SkipResult(next_ts=self.state.window_ts)

        min_speech_duration_ms = 750
        min_speech_duration_samples = 16000 * min_speech_duration_ms // 1000
        audio_samples = self.state.audio_ts - self.state.window_ts
        voice_segments = self.vad_callback(self.state.rolling_window.array[0:audio_samples], min_speech_duration_ms, 100)
        LOG.debug(f"Voice segments: {voice_segments}")
        if len(voice_segments) == 0:
            # preserve a bit of audio at the end that may have been too short just because it is truncated
            padding = min_speech_duration_samples
            remaining_audio = self.state.audio_ts - self.state.window_ts - padding
            advance_to = remaining_audio
            # pylint: disable=consider-using-min-builtin
            if max_skip < advance_to:
                advance_to = max_skip
            LOG.debug(
                f"Skipping silences between {samples_to_seconds(self.state.window_ts):.2f}"
                f" and {samples_to_seconds(self.state.window_ts + advance_to):.2f}"
            )
            marker = self._build_skip_marker(
                start_ts=self.state.window_ts,
                end_ts=self.state.window_ts + advance_to,
                samples=self.state.rolling_window.array[0:advance_to],
            )
            self.state.advance_audio_window(advance_to)
            return SkipResult(next_ts=self.state.audio_ts, marker=marker)

        voice_start = voice_segments[0]["start"]
        voice_end = voice_segments[0]["end"]
        voice_length = voice_end - voice_start

        # skip leading silences
        if voice_start > 0:
            # pylint: disable=consider-using-min-builtin
            advance_to = voice_start
            if max_skip < advance_to:
                advance_to = max_skip
            LOG.debug(
                f"Skipping silences between {samples_to_seconds(self.state.window_ts):.2f}"
                f" and {samples_to_seconds(self.state.window_ts + advance_to):.2f}"
            )
            marker = self._build_skip_marker(
                start_ts=self.state.window_ts,
                end_ts=self.state.window_ts + advance_to,
                samples=self.state.rolling_window.array[0:advance_to],
            )
            self.state.advance_audio_window(advance_to)
            return SkipResult(next_ts=voice_length + self.state.window_ts, marker=marker)

        return SkipResult(next_ts=voice_length + self.state.window_ts)

    def dump_window_to_file(self, filename: str = "debug_window.wav"):
        window = self.state.rolling_window.array
        LOG.debug("Dumping current window to file for debugging.")
        # pylint: disable=no-member
        with wave.open(filename, "wb") as wf:
            wf.setnchannels(1)  # Mono
            wf.setsampwidth(2)  # 16-bit audio
            wf.setframerate(self.config.sampling_rate)
            wf.writeframes((window * 32768.0).astype(np.int16).tobytes())
        LOG.debug("Finished writing window to file.")

    @staticmethod
    def align_words_to_sentences(id_provider: IdProvider, sentences: list[str], window_words: list[Word]) -> list[Utterance]:
        # Concatenate all sentences into a single string
        full_text_sentences = "".join(sentences)
        # Concatenate all words from window_words
        full_text_words = "".join(w.word for w in window_words)

        # Basic validation step (optional but recommended)
        if full_text_sentences != full_text_words:
            raise ValueError("The joined text from sentences and window_words do not match.")

        # We'll iterate through sentences and pick words until we match them character-for-character.
        result = []
        current_char_index = 0
        current_word_index = 0
        word_char_offset = 0

        def remove_spaces_and_punctuation(string: str) -> str:
            return string.translate(str.maketrans("", "", " " + PREPEND_PUNCTUATIONS + APPEND_PUNCTUATIONS))

        # For each sentence, we want to collect a sublist of VerbatimWord
        for sentence_index, sentence in enumerate(sentences):
            raw_sentence = sentence
            # igonre punctuation and spaces
            sentence = remove_spaces_and_punctuation(sentence)
            sentence_length = len(sentence.strip())
            target_end = current_char_index + sentence_length

            sentence_words = []
            # Accumulate words until we have accounted for all chars of this sentence
            while current_char_index < target_end:
                if current_word_index >= len(window_words):
                    raise ValueError("Ran out of words while aligning sentences.")

                w = window_words[current_word_index]
                word_text = remove_spaces_and_punctuation(w.word)
                word_length = len(word_text)

                # Words that resolve to only punctuation/whitespace carry no characters; assign them directly.
                if word_length == 0:
                    sentence_words.append(w)
                    current_word_index += 1
                    word_char_offset = 0
                    continue

                remaining_word_chars = word_length - word_char_offset
                if remaining_word_chars <= 0:
                    # Safety: reset offset if we somehow consumed the full word previously.
                    word_char_offset = 0
                    continue

                remaining_sentence_chars = target_end - current_char_index
                if remaining_sentence_chars <= 0:
                    break

                if remaining_word_chars <= remaining_sentence_chars:
                    # Entire remaining portion of the word fits in this sentence.
                    current_char_index += remaining_word_chars
                    sentence_words.append(w)
                    current_word_index += 1
                    word_char_offset = 0
                else:
                    # Only part of the word belongs to this sentence; defer adding the word until it is fully consumed.
                    LOG.debug(
                        "Sentence alignment split inside word: sentence_index=%d raw=%r normalized=%r target_end=%d "
                        "word=%r word_range=%d-%d word_char_offset=%d remaining_word_chars=%d remaining_sentence_chars=%d",
                        sentence_index,
                        raw_sentence,
                        sentence,
                        target_end,
                        w.word,
                        w.start_ts,
                        w.end_ts,
                        word_char_offset,
                        remaining_word_chars,
                        remaining_sentence_chars,
                    )
                    current_char_index += remaining_sentence_chars
                    word_char_offset += remaining_sentence_chars
                    break

            # Now we have all the words for this sentence
            if len(sentence_words) > 0:
                utterance = Utterance.from_words(utterance_id=id_provider.next(), words=sentence_words)
                LOG.debug(
                    "Sentence alignment result: sentence_index=%d raw=%r normalized=%r utterance=%r words=%s",
                    sentence_index,
                    raw_sentence,
                    sentence,
                    utterance.text,
                    [word.word for word in sentence_words],
                )
                result.append(utterance)
            else:
                LOG.debug(
                    "Sentence alignment produced empty utterance: sentence_index=%d raw=%r normalized=%r target_end=%d",
                    sentence_index,
                    raw_sentence,
                    sentence,
                    target_end,
                )

        # At the end, `result` should be a List[List[VerbatimWord]]
        return result

    @staticmethod
    def words_to_sentences(word_tokenizer: SentenceTokenizer, window_words: List[Word], id_provider: IdProvider) -> list[Utterance]:
        sentences = []
        if len(window_words) == 0:
            return []

        for tok in word_tokenizer.split(words=window_words):
            sentences += [tok]

        LOG.debug(
            "Sentence tokenizer output: tokenizer=%s words=%s sentences=%s",
            type(word_tokenizer).__name__,
            [word.word for word in window_words],
            sentences,
        )

        utterances = Verbatim.align_words_to_sentences(id_provider=id_provider, sentences=sentences, window_words=window_words)
        return utterances

    def guess_language(self, timestamp: int) -> Tuple[str, float, int]:
        request = LanguageDetectionRequest(
            audio=self.state.rolling_window.array,
            lang=self.config.lang,
            timestamp=timestamp,
            window_ts=self.state.window_ts,
            audio_ts=self.state.audio_ts,
        )
        result: LanguageDetectionResult = detect_language(request=request, guess_fn=self.language_identifier.guess_language)
        return (result.language, result.probability, result.samples_used)

    def transcribe_window(self) -> Tuple[List[Word], List[Word]]:
        LOG.debug("Starting transcription of audio chunk.")
        LOG.debug("Window Start Time: %s", samples_to_seconds(self.state.window_ts))
        LOG.debug("Confirmed Time: %s", samples_to_seconds(self.state.confirmed_ts))
        LOG.debug("Acknowledged Time: %s", samples_to_seconds(self.state.acknowledged_ts))
        LOG.debug("Valid audio range: 0.0 - %s", samples_to_seconds(self.state.audio_ts - self.state.window_ts))

        acknowledged_words_in_window = WhisperHistory.advance_transcript(timestamp=self.state.window_ts, transcript=self.state.acknowledged_words)
        detect_start = perf_counter()
        lang, _prob, used_samples_for_language = self.guess_language(timestamp=max(0, self.state.acknowledged_ts))
        self.state.last_transcribe_lang = lang
        detect_ms = (perf_counter() - detect_start) * 1000.0
        prefix_text = ""
        for word in [w for u in self.state.unacknowledged_utterances for w in u.words]:
            if word.lang != lang:
                break
            prefix_text += word.word
        whisper_prompt = self.config.whisper_prompts[lang] if lang in self.config.whisper_prompts else self.config.whisper_prompts["en"]

        try:
            transcribe_start = perf_counter()
            transcript_words = self.models.transcriber.transcribe(
                audio=self.state.rolling_window.array[0 : self.state.audio_ts - self.state.window_ts],
                lang=lang,
                prompt=whisper_prompt,
                prefix=prefix_text,
                window_ts=self.state.window_ts,
                audio_ts=self.state.audio_ts,
                whisper_beam_size=self.config.whisper_beam_size,
                whisper_best_of=self.config.whisper_best_of,
                whisper_patience=self.config.whisper_patience,
                whisper_temperatures=self.config.whisper_temperatures,
            )
            transcribe_ms = (perf_counter() - transcribe_start) * 1000.0
        except RuntimeError as e:
            LOG.warning(f"Transcription failed with RuntimeError: {str(e)}. Skipping this chunk.")
            self.state.last_transcribe_metrics_ms = {
                "detect_ms": detect_ms,
                "transcribe_ms": 0.0,
                "confirm_ms": 0.0,
            }
            return [], []

        if len(transcript_words) > 0 and len(self.config.lang) > 1:
            # make sure that the first word detected is within the audio samples used
            # to guess the language; otherwise, language might have been guess from a short utterance such
            # as "hmm hmm" followed by a pause and then actual words

            if transcript_words[0].start_ts >= self.state.window_ts + used_samples_for_language:
                # So the words detected are not part of the samples evaluated when guessing the language
                # in this unfortunate situation, we could simply perform language detected on the range of
                # words detected, but there is an non-zero chance that the transcription skipped words
                # because we suggested the wrong langauge. So instead, we try the following:
                # 1. Transcribe in each of the supported language;
                # 2. Confirm with language detection that we used the appropriate language
                # 3. if language detection does not match, ignore the transcripton
                # 4. within the transcription that matched, process the one that starts with the earliest timestamp

                best_first_word_ts = self.state.audio_ts
                best_transcript = transcript_words
                best_lang = lang
                for test_lang in self.config.lang:
                    if test_lang == lang:
                        first_word_ts = transcript_words[0].start_ts
                        confirmed_lang, _prob, _used_samples_for_language = self.guess_language(timestamp=max(0, first_word_ts))
                        if confirmed_lang == test_lang:
                            if transcript_words[0].start_ts < best_first_word_ts:
                                best_first_word_ts = transcript_words[0].start_ts
                                best_transcript = transcript_words
                                best_lang = lang
                    else:
                        alt_prefix_text = ""
                        for word in [w for u in self.state.unacknowledged_utterances for w in u.words]:
                            if word.lang != test_lang:
                                break
                            alt_prefix_text += word.word
                        alt_whisper_prompt = (
                            self.config.whisper_prompts[test_lang] if test_lang in self.config.whisper_prompts else self.config.whisper_prompts["en"]
                        )
                        alt_transcribe_start = perf_counter()
                        alt_transcript_words = self.models.transcriber.transcribe(
                            audio=self.state.rolling_window.array[0 : self.state.audio_ts - self.state.window_ts],
                            lang=test_lang,
                            prompt=alt_whisper_prompt,
                            prefix=alt_prefix_text,
                            window_ts=self.state.window_ts,
                            audio_ts=self.state.audio_ts,
                            whisper_beam_size=self.config.whisper_beam_size,
                            whisper_best_of=self.config.whisper_best_of,
                            whisper_patience=self.config.whisper_patience,
                            whisper_temperatures=self.config.whisper_temperatures,
                        )
                        transcribe_ms += (perf_counter() - alt_transcribe_start) * 1000.0
                        if len(alt_transcript_words) > 0:
                            first_word_ts = alt_transcript_words[0].start_ts
                            confirmed_lang, _prob, _used_samples_for_language = self.guess_language(timestamp=max(0, first_word_ts))
                            if confirmed_lang == test_lang:
                                if alt_transcript_words[0].start_ts < best_first_word_ts:
                                    best_first_word_ts = alt_transcript_words[0].start_ts
                                    best_transcript = alt_transcript_words
                                    best_lang = test_lang
                lang = best_lang
                transcript_words = best_transcript

        confirm_start = perf_counter()
        self.state.transcript_candidate_history.advance(self.state.window_ts)
        confirmed_words = self.state.transcript_candidate_history.confirm(
            current_words=transcript_words,
            after_ts=self.state.acknowledged_ts - 1,
            prefix=acknowledged_words_in_window,
        )
        self.state.transcript_candidate_history.add(transcription=transcript_words)

        if len(confirmed_words) > 0:
            self.state.confirmed_ts = confirmed_words[-1].start_ts
            LOG.debug(f"Confirmed ts: {self.state.confirmed_ts} ({samples_to_seconds(self.state.confirmed_ts)})")

            transcript_history = self.state.transcript_candidate_history.transcript_history
            transcript_lines = [
                f"H{index}: {self._summarize_words(history)} text={self._words_text(history)!r}"
                for index, history in enumerate(transcript_history)
            ]
            transcript_lines += [
                f"CUR: {self._summarize_words(transcript_words)} text={self._words_text(transcript_words)!r}",
                f"CONF: {self._summarize_words(confirmed_words)} text={self._words_text(confirmed_words)!r}",
            ]
            LOG.debug(
                "Transcript candidates:\n%s",
                os.linesep.join(transcript_lines),
            )
            LOG.debug(
                "Transcript summary: lang=%s current=%s confirmed=%s history=%s",
                self.state.last_transcribe_lang,
                self._summarize_words(transcript_words),
                self._summarize_words(confirmed_words),
                " | ".join(self._summarize_words(history) for history in transcript_history),
            )
        confirm_ms = (perf_counter() - confirm_start) * 1000.0

        unconfirmed_words = [transcript_words[i] for i in range(len(confirmed_words), len(transcript_words))]
        self.state.last_transcribe_metrics_ms = {
            "detect_ms": detect_ms,
            "transcribe_ms": transcribe_ms,
            "confirm_ms": confirm_ms,
        }
        return confirmed_words, unconfirmed_words

    def gap_language_mismatch(self, *, start_ts: int, end_ts: int, expected_lang: str, min_gap_samples: int = 4000) -> bool:
        if not expected_lang or end_ts <= start_ts:
            return False
        if end_ts - start_ts < min_gap_samples:
            return False

        probe_ts = start_ts + (end_ts - start_ts) // 2
        detected_lang, probability, _used_samples = self.guess_language(timestamp=max(0, probe_ts))
        probe_start = max(self.state.window_ts, probe_ts - 8000)
        probe_end = min(self.state.audio_ts, probe_ts + 8000)
        nearby_unconfirmed = [word for word in self.state.unconfirmed_words if word.end_ts >= probe_start and word.start_ts <= probe_end]
        mismatch = probability > 0.5 and detected_lang != expected_lang
        LOG.debug(
            "Gap probe %d-%d expected=%s detected=%s prob=%.3f mismatch=%s probe_ts=%d nearby_unconfirmed=%s",
            start_ts,
            end_ts,
            expected_lang,
            detected_lang,
            probability,
            mismatch,
            probe_ts,
            self._summarize_words(nearby_unconfirmed),
        )
        return mismatch

    def get_next_number_of_chunks(self) -> int:
        available_chunks = self.config.window_duration - float(self.state.audio_ts - self.state.window_ts) / self.config.sampling_rate

        thresholds = self.config.chunk_table

        for limit, value in thresholds:
            limit_sample = int(limit * self.config.window_duration)
            value_sample = math.ceil(value * self.config.window_duration)
            if available_chunks >= limit_sample:
                return value_sample

        # If for some reason available_chunks is less than 0, return the smallest chunk count.
        return 1

    def pretty_print_transcript(
        self,
        acknowledged_utterances: List[Utterance],
        unacknowledged_utterances: List[Utterance],
        unconfirmed_words: List[Word],
        file: TextIO = sys.stdout,
    ):
        colour_scheme_ack = COLORSCHEME_ACKNOWLEDGED if self.config.log_colours else COLORSCHEME_NONE
        colour_scheme_unack = COLORSCHEME_UNACKNOWLEDGED if self.config.log_colours else COLORSCHEME_NONE
        colour_scheme_unconfirmed = COLORSCHEME_UNCONFIRMED if self.config.log_colours else COLORSCHEME_NONE
        formatter: TranscriptFormatter = TranscriptFormatter(
            speaker_style=SpeakerStyle.always,
            timestamp_style=TimestampStyle.range,
            probability_style=ProbabilityStyle.word,
            language_style=LanguageStyle.always,
        )
        prefix = (
            f"[{samples_to_seconds(self.state.window_ts)}/"
            f"{samples_to_seconds(self.state.audio_ts - self.state.acknowledged_ts)}/"
            f"{samples_to_seconds(self.state.audio_ts - self.state.confirmed_ts)}]"
        )
        file.write(prefix + (Fore.LIGHTGREEN_EX if self.config.log_colours else ""))
        for u in acknowledged_utterances:
            file.write(formatter.format_utterance(utterance=u, colours=colour_scheme_ack).decode("utf-8"))
        for u in unacknowledged_utterances:
            file.write(formatter.format_utterance(utterance=u, colours=colour_scheme_unack).decode("utf-8"))
        if len(unconfirmed_words) > 0:
            file.write(
                formatter.format_utterance(
                    utterance=Utterance.from_words(utterance_id=self.state.utterance_id.next(), words=unconfirmed_words),
                    colours=colour_scheme_unconfirmed,
                ).decode("utf-8")
            )
        file.write(os.linesep)
        file.flush()

    def acknowledge_utterances(
        self,
        utterances: List[Utterance],
        min_ack_duration=16000,
        min_unack_duration=16000,
    ) -> Tuple[List[Utterance], List[Utterance]]:
        if len(utterances) == 0:
            return [], utterances

        # check if the last utterance is complete
        last_valid_utterance_index = len(utterances) - 1
        valid_endings = tuple(APPEND_PUNCTUATIONS)
        if not utterances[last_valid_utterance_index].text.endswith(valid_endings):
            last_valid_utterance_index = last_valid_utterance_index - 1

        if last_valid_utterance_index < 0:
            return [], utterances

        start_ts = utterances[0].start_ts
        max_duration = self.state.audio_ts - self.state.window_ts - min_unack_duration

        duration = 0
        index = 0
        while duration < min_ack_duration and index <= last_valid_utterance_index:
            new_duration = utterances[index].end_ts - start_ts
            if new_duration > max_duration:
                break
            duration = new_duration
            index += 1

        return utterances[0:index], utterances[index:]

    def get_speaker_at(self, time: float, diarization: Optional[Annotation]) -> Optional[str]:
        if diarization is None:
            return None

        for turn, _track, speaker in diarization.itertracks(yield_label=True):  # pyright: ignore[reportAssignmentType]
            if turn.end < time:
                continue
            if turn.start > time:
                break
            return str(speaker)
        return None

    def get_speaker_before(self, time: float, diarization: Annotation) -> Optional[str]:
        last_speaker = None
        for turn, _track, speaker in diarization.itertracks(yield_label=True):  # pyright: ignore[reportAssignmentType]
            if turn.end < time:
                last_speaker = speaker
                continue
            if turn.start > time:
                break
        return str(last_speaker)

    def get_speaker_after(self, time: float, diarization: Annotation) -> Optional[str]:
        for turn, _track, speaker in diarization.itertracks(yield_label=True):  # pyright: ignore[reportAssignmentType]
            if turn.start > time:
                return str(speaker)
        return None

    def assign_speaker(self, utterance: Utterance, diarization: Optional[Annotation]) -> Optional[str]:
        if diarization is None:
            return None

        start = utterance.start_ts / 16000.0
        end = utterance.end_ts / 16000.0
        duration = end - start
        samples: List[float]
        if duration < 1:
            samples = [start + duration * 0.5]
        elif duration < 4:
            samples = [
                start + duration * 0.25,
                start + duration * 0.5,
                start + duration * 0.75,
            ]
        else:
            samples = [
                start + duration * 0.2,
                start + duration * 0.3,
                start + duration * 0.4,
                start + duration * 0.5,
                start + duration * 0.6,
                start + duration * 0.7,
                start + duration * 0.8,
            ]

        votes = {}
        speaker: Optional[str]
        for sample in samples:
            speaker: Optional[str] = self.get_speaker_at(time=sample, diarization=diarization)
            if speaker:
                if speaker in votes:
                    votes[speaker] = votes[speaker] + 1
                else:
                    votes[speaker] = 1

        if len(votes) != 0:
            best_votes = 0
            best_speaker = ""
            for speaker, vote in votes.items():
                if vote > best_votes:
                    best_speaker = speaker
                    best_votes = vote
            return best_speaker

        speaker = self.get_speaker_before(time=start + duration * 0.5, diarization=diarization)
        if speaker:
            return speaker

        speaker = self.get_speaker_after(time=start + duration * 0.5, diarization=diarization)
        if speaker:
            return speaker

        return None

    def process_audio_window(self, audio_stream: AudioStream) -> Generator[Tuple[Utterance, List[Utterance], List[Word]], None, None]:
        while True:
            iteration_start = perf_counter()
            initial_window_ts = self.state.window_ts
            initial_confirmed_ts = self.state.confirmed_ts
            initial_acknowledged_ts = self.state.acknowledged_ts
            # minimum number of samples to attempt transcription
            min_audio_duration_samples = 16000
            min_speech_duration_ms = 500
            utterances = []
            pass_metrics = {
                "detect_ms": 0.0,
                "transcribe_ms": 0.0,
                "confirm_ms": 0.0,
                "sentence_ms": 0.0,
                "acknowledge_ms": 0.0,
                "speaker_ms": 0.0,
                "output_ms": 0.0,
                "advance_ms": 0.0,
            }
            enable_vad = True
            if enable_vad and self.state.skip_silences:
                self.state.skip_silences = False
                next_ts = self.state.audio_ts
                if len(self.state.unacknowledged_utterances) > 0:
                    next_ts = min(next_ts, self.state.unacknowledged_utterances[0].start_ts)
                if len(self.state.unconfirmed_words) > 0:
                    next_ts = min(next_ts, self.state.unconfirmed_words[0].start_ts)
                skip_result = self.skip_leading_silence(min_speech_duration_ms=min_speech_duration_ms, max_skip=next_ts - self.state.window_ts)
                if skip_result.marker is not None:
                    yield skip_result.marker, [], []
                if self.state.audio_ts - self.state.window_ts < min_audio_duration_samples:
                    # we skipped all available audio - keep skipping silences and do nothing else for now
                    self.state.skip_silences = True

            if self.state.audio_ts - self.state.window_ts < min_audio_duration_samples:
                return

            if self.config.debug and self.config.working_dir:
                self.dump_window_to_file(filename=f"{self.state.working_prefix_no_ext}-debug_window.wav")  # Dump current window for debugging

            confirmed_words, unconfirmed_words = self.transcribe_window()
            pass_metrics.update(self.state.last_transcribe_metrics_ms)
            self.state.unconfirmed_words = unconfirmed_words
            if len(confirmed_words) > 0:
                sentence_start = perf_counter()
                utterances = self.words_to_sentences(
                    word_tokenizer=BoundedSentenceTokenizer(other_tokenizer=self.models.sentence_tokenizer),
                    window_words=confirmed_words,
                    id_provider=self.state.utterance_id,
                )
                pass_metrics["sentence_ms"] += (perf_counter() - sentence_start) * 1000.0

                acknowledge_start = perf_counter()
                acknowledged_utterances, confirmed_utterances = self.acknowledge_utterances(utterances=utterances)
                pass_metrics["acknowledge_ms"] += (perf_counter() - acknowledge_start) * 1000.0

                acknowledged_to_commit = acknowledged_utterances[:1]
                deferred_acknowledged = acknowledged_utterances[1:]
                LOG.debug(
                    "Ack candidates: lang=%s anchor=%d acknowledged=%s commit=%s defer=%s confirmed=%s unconfirmed=%s",
                    self.state.last_transcribe_lang,
                    max(self.state.window_ts, self.state.acknowledged_ts),
                    self._summarize_utterances(acknowledged_utterances),
                    self._summarize_utterances(acknowledged_to_commit),
                    self._summarize_utterances(deferred_acknowledged),
                    self._summarize_utterances(confirmed_utterances),
                    self._summarize_words(unconfirmed_words),
                )

                if len(acknowledged_to_commit) > 0:
                    first_candidate = acknowledged_to_commit[0]
                    current_anchor_ts = max(self.state.window_ts, self.state.acknowledged_ts)
                    if self.gap_language_mismatch(
                        start_ts=current_anchor_ts,
                        end_ts=first_candidate.start_ts,
                        expected_lang=self.state.last_transcribe_lang,
                    ):
                        conservative_skip_samples = 100 * 16000 // 1000
                        previous_pending = list(self.state.unacknowledged_utterances)
                        self.state.unacknowledged_utterances = deferred_acknowledged + confirmed_utterances
                        if conservative_skip_samples > 0:
                            self.state.advance_audio_window(conservative_skip_samples)
                        self.state.skip_silences = True
                        LOG.debug(
                            "Blocked first candidate due to gap mismatch: lang=%s anchor=%d candidate=%s prev_pending=%s new_pending=%s",
                            self.state.last_transcribe_lang,
                            current_anchor_ts,
                            self._summarize_utterance(first_candidate),
                            self._summarize_utterances(previous_pending),
                            self._summarize_utterances(self.state.unacknowledged_utterances),
                        )
                        break

                if audio_stream.diarization:
                    speaker_start = perf_counter()
                    for acknowledged_utterance in acknowledged_to_commit:
                        acknowledged_utterance.speaker = self.assign_speaker(acknowledged_utterance, audio_stream.diarization)
                    pass_metrics["speaker_ms"] += (perf_counter() - speaker_start) * 1000.0

                previous_pending = list(self.state.unacknowledged_utterances)
                self.state.acknowledged_utterances += acknowledged_to_commit
                self.state.unacknowledged_utterances = deferred_acknowledged + confirmed_utterances
                LOG.debug(
                    "Pending replace: lang=%s prev=%s new=%s",
                    self.state.last_transcribe_lang,
                    self._summarize_utterances(previous_pending),
                    self._summarize_utterances(self.state.unacknowledged_utterances),
                )

                for i, utterance in enumerate(acknowledged_to_commit):
                    result: TranscriptionWindowResult[Utterance, Word] = TranscriptionWindowResult(
                        utterance=utterance,
                        unacknowledged=acknowledged_to_commit[i + 1 :] + deferred_acknowledged + confirmed_utterances,
                        unconfirmed_words=unconfirmed_words,
                    )
                    yield result.as_tuple()

                if len(acknowledged_to_commit) > 0:
                    for u in acknowledged_to_commit:
                        self.state.acknowledged_words += u.words

                    # utterances are split at short pauses, advance a bit to avoid repeating the last word
                    # but not too much as to skip the first word of the next utterance
                    utterance_padding_ms = 100
                    utterance_padding_samples = utterance_padding_ms * 16000 // 1000
                    accepted_end_ts = acknowledged_to_commit[-1].end_ts
                    skip_to = accepted_end_ts + utterance_padding_samples

                    next_visible_ts = None
                    if len(self.state.unacknowledged_utterances) > 0:
                        next_visible_ts = self.state.unacknowledged_utterances[0].start_ts
                    if len(self.state.unconfirmed_words) > 0:
                        next_unconfirmed_ts = self.state.unconfirmed_words[0].start_ts
                        if next_visible_ts is None or next_unconfirmed_ts < next_visible_ts:
                            next_visible_ts = next_unconfirmed_ts

                    if next_visible_ts is not None and self.gap_language_mismatch(
                        start_ts=accepted_end_ts,
                        end_ts=next_visible_ts,
                        expected_lang=self.state.last_transcribe_lang,
                    ):
                        skip_to = accepted_end_ts
                        LOG.debug(
                            "Suppressing post-ack padding due to gap mismatch: lang=%s accepted=%s next_visible=%d",
                            self.state.last_transcribe_lang,
                            self._summarize_utterance(acknowledged_to_commit[-1]),
                            next_visible_ts,
                        )

                    # try to skip ahead, but don't skip beyond the next detected word
                    if len(self.state.unacknowledged_utterances) > 0 and skip_to > self.state.unacknowledged_utterances[0].start_ts:
                        skip_to = self.state.unacknowledged_utterances[0].start_ts
                    if len(self.state.unconfirmed_words) > 0 and skip_to > self.state.unconfirmed_words[0].start_ts:
                        skip_to = self.state.unconfirmed_words[0].start_ts

                    # Never acknowledge beyond audio we've actually ingested
                    self.state.acknowledged_ts = min(skip_to, self.state.audio_ts)
                    self.state.skip_silences = True
                    LOG.debug(
                        "Committed utterance: lang=%s utterance=%s next_pending=%s next_unconfirmed=%s acknowledged_ts=%d",
                        self.state.last_transcribe_lang,
                        self._summarize_utterance(acknowledged_to_commit[-1]),
                        self._summarize_utterances(self.state.unacknowledged_utterances),
                        self._summarize_words(self.state.unconfirmed_words),
                        self.state.acknowledged_ts,
                    )
            else:
                acknowledged_utterances = []
                confirmed_utterances = []

            output_start = perf_counter()
            outstr = StringIO()
            self.pretty_print_transcript(
                acknowledged_utterances=[],
                unacknowledged_utterances=confirmed_utterances,
                unconfirmed_words=unconfirmed_words,
                file=outstr,
            )
            LOG.debug(outstr.getvalue())
            pass_metrics["output_ms"] += (perf_counter() - output_start) * 1000.0

            self.state.acknowledged_words = WhisperHistory.advance_transcript(
                timestamp=self.state.window_ts, transcript=self.state.acknowledged_words
            )

            if self.state.acknowledged_ts > self.state.window_ts:
                advance_start = perf_counter()
                shift_amount = self.state.acknowledged_ts - self.state.window_ts
                self.state.advance_audio_window(shift_amount)
                self.state.skip_silences = True
                pass_metrics["advance_ms"] += (perf_counter() - advance_start) * 1000.0

            progressed_samples = max(0, self.state.window_ts - initial_window_ts)
            confirmed_delta_s = samples_to_seconds(max(0, self.state.confirmed_ts - initial_confirmed_ts))
            acknowledged_delta_s = samples_to_seconds(max(0, self.state.acknowledged_ts - initial_acknowledged_ts))
            total_pass_ms = (perf_counter() - iteration_start) * 1000.0
            pass_metrics["total_ms"] = total_pass_ms
            self.state.timing_totals_ms["detect"] += pass_metrics["detect_ms"]
            self.state.timing_totals_ms["transcribe"] += pass_metrics["transcribe_ms"]
            self.state.timing_totals_ms["confirm"] += pass_metrics["confirm_ms"]
            self.state.timing_totals_ms["sentence"] += pass_metrics["sentence_ms"]
            self.state.timing_totals_ms["acknowledge"] += pass_metrics["acknowledge_ms"]
            self.state.timing_totals_ms["speaker"] += pass_metrics["speaker_ms"]
            self.state.timing_totals_ms["output"] += pass_metrics["output_ms"]
            self.state.timing_totals_ms["advance"] += pass_metrics["advance_ms"]
            self.state.timing_totals_ms["total"] += pass_metrics["total_ms"]
            LOG.info(
                (
                    "Pass timing: total=%.1fms detect=%.1fms transcribe=%.1fms confirm=%.1fms "
                    "sentence=%.1fms acknowledge=%.1fms speaker=%.1fms output=%.1fms advance=%.1fms "
                    "confirmed_delta=%.3fs ack_delta=%.3fs"
                ),
                pass_metrics["total_ms"],
                pass_metrics["detect_ms"],
                pass_metrics["transcribe_ms"],
                pass_metrics["confirm_ms"],
                pass_metrics["sentence_ms"],
                pass_metrics["acknowledge_ms"],
                pass_metrics["speaker_ms"],
                pass_metrics["output_ms"],
                pass_metrics["advance_ms"],
                confirmed_delta_s,
                acknowledged_delta_s,
            )
            if progressed_samples > 0:
                iteration_elapsed_seconds = max(perf_counter() - iteration_start, 1e-9)
                iteration_progress_seconds = samples_to_seconds(progressed_samples)
                iteration_speed_multiplier = iteration_progress_seconds / iteration_elapsed_seconds
                total_elapsed_seconds = max(perf_counter() - self.state.processing_started_at, 1e-9)
                total_progressed_samples = max(0, self.state.window_ts - self.state.processing_started_window_ts)
                total_progress_seconds = samples_to_seconds(total_progressed_samples)
                average_speed_multiplier = total_progress_seconds / total_elapsed_seconds
                total_samples = self._get_total_samples(audio_stream)
                current_seconds = samples_to_seconds(self.state.window_ts)
                if total_samples is not None and total_samples > 0:
                    total_seconds = samples_to_seconds(total_samples)
                    progress_percent = 100.0 * self.state.window_ts / total_samples
                    LOG.info(
                        "Clip progress %.3fs/%.3fs (%.1f%%) speed=%.2fx avg=%.2fx",
                        current_seconds,
                        total_seconds,
                        progress_percent,
                        iteration_speed_multiplier,
                        average_speed_multiplier,
                    )
                else:
                    LOG.info(
                        "Clip progress %.3fs speed=%.2fx avg=%.2fx",
                        current_seconds,
                        iteration_speed_multiplier,
                        average_speed_multiplier,
                    )

            if len(acknowledged_utterances) <= 1:
                break

    def capture_audio(self, audio_source: AudioStream):
        if not audio_source.has_more():
            return False

        next_chunk = self.get_next_number_of_chunks()
        LOG.debug(f"capturing {next_chunk} chunks of audio")
        audio_array = audio_source.next_chunk(next_chunk)
        self.state.append_audio_to_window(audio_array)
        return True

    def flush_overflowing_utterances(self, diarization: Optional[Annotation]) -> Generator[Tuple[Utterance, List[Utterance], List[Word]], None, None]:
        # As the attention window advances, we may not be able to acknowledge
        # all utterances and words; When they fall behind, the best we can do
        # is return them as acknowledge.

        # First, try to acknowledge "full" utterances that may not have been acknowledged
        while len(self.state.unacknowledged_utterances) > 0:
            # Stop when we reached the window_ts
            # Note that the == case is frequent, whisper can generate several words with timestamp 0 -> 0
            if self.state.unacknowledged_utterances[0].end_ts >= self.state.window_ts:
                break
            utterance = self.state.unacknowledged_utterances.pop(0)
            utterance.speaker = self.assign_speaker(utterance, diarization)
            self.state.acknowledged_words += utterance.words
            result: TranscriptionWindowResult[Utterance, Word] = TranscriptionWindowResult(
                utterance=utterance,
                unacknowledged=self.state.unacknowledged_utterances,
                unconfirmed_words=self.state.unconfirmed_words,
            )
            yield result.as_tuple()

        # If an utterance crosses the boundary, acknowledge the whole utterance so the
        # next pass can restart from a clean utterance boundary instead of mid-utterance.
        if len(self.state.unacknowledged_utterances) > 0:
            crossing_utterance = self.state.unacknowledged_utterances[0]
            if crossing_utterance.start_ts < self.state.window_ts:
                utterance = self.state.unacknowledged_utterances.pop(0)
                utterance.speaker = self.assign_speaker(utterance, diarization)
                self.state.acknowledged_words += utterance.words
                result: TranscriptionWindowResult[Utterance, Word] = TranscriptionWindowResult(
                    utterance=utterance,
                    unacknowledged=self.state.unacknowledged_utterances,
                    unconfirmed_words=self.state.unconfirmed_words,
                )
                yield result.as_tuple()

        # If there are unconfirmed words falling behind, acknowledge them
        flushed_utterances_words = []
        while len(self.state.unconfirmed_words) > 0:
            # Stop when we reached the window_ts
            # Note that the == case is frequent, whisper can generate several words with timestamp 0 -> 0
            if self.state.unconfirmed_words[0].end_ts >= self.state.window_ts:
                break
            flushed_word = self.state.unconfirmed_words.pop(0)
            flushed_utterances_words.append(flushed_word)

        if len(flushed_utterances_words) > 0:
            utterance = Utterance.from_words(utterance_id=self.state.utterance_id.next(), words=flushed_utterances_words)
            utterance.speaker = self.assign_speaker(utterance, diarization)
            self.state.acknowledged_words += flushed_utterances_words
            result: TranscriptionWindowResult[Utterance, Word] = TranscriptionWindowResult(
                utterance=utterance,
                unacknowledged=self.state.unacknowledged_utterances,
                unconfirmed_words=self.state.unconfirmed_words,
            )
            yield result.as_tuple()

    def transcribe_naive(
        self, audio_stream: AudioStream, working_prefix_no_ext: str = "out"
    ) -> Generator[Tuple[Utterance, List[Utterance], List[Word]], None, None]:
        self.state = State(self.config, working_prefix_no_ext=working_prefix_no_ext)
        if audio_stream.start_offset != 0:
            self.state.window_ts = audio_stream.start_offset
            self.state.audio_ts = audio_stream.start_offset

        self._emit_status(StatusUpdate(state="transcribing"))
        full_audio = self._read_full_audio(audio_stream)
        if full_audio.size == 0:
            return

        self.state.audio_ts = self.state.window_ts + len(full_audio)

        detect_start = perf_counter()
        lang_result = self._detect_language_for_audio(
            audio=full_audio,
            window_ts=self.state.window_ts,
            audio_ts=self.state.audio_ts,
            timestamp=max(0, self.state.window_ts),
        )
        lang = lang_result.language
        self.state.last_transcribe_lang = lang
        detect_ms = (perf_counter() - detect_start) * 1000.0

        whisper_prompt = self.config.whisper_prompts[lang] if lang in self.config.whisper_prompts else self.config.whisper_prompts["en"]
        transcribe_start = perf_counter()
        transcript_words = self.models.transcriber.transcribe(
            audio=full_audio,
            lang=lang,
            prompt=whisper_prompt,
            prefix="",
            window_ts=self.state.window_ts,
            audio_ts=self.state.audio_ts,
            whisper_beam_size=self.config.whisper_beam_size,
            whisper_best_of=self.config.whisper_best_of,
            whisper_patience=self.config.whisper_patience,
            whisper_temperatures=self.config.whisper_temperatures,
        )
        transcribe_ms = (perf_counter() - transcribe_start) * 1000.0
        self.state.timing_totals_ms["detect"] += detect_ms
        self.state.timing_totals_ms["transcribe"] += transcribe_ms

        if len(transcript_words) == 0:
            self.state.timing_totals_ms["total"] += detect_ms + transcribe_ms
            return

        sentence_start = perf_counter()
        utterances = self.words_to_sentences(
            word_tokenizer=self.models.sentence_tokenizer,
            window_words=transcript_words,
            id_provider=self.state.utterance_id,
        )
        sentence_ms = (perf_counter() - sentence_start) * 1000.0
        self.state.timing_totals_ms["sentence"] += sentence_ms

        speaker_ms = 0.0
        if audio_stream.diarization:
            speaker_start = perf_counter()
            for utterance in utterances:
                utterance.speaker = self.assign_speaker(utterance, audio_stream.diarization)
            speaker_ms = (perf_counter() - speaker_start) * 1000.0
            self.state.timing_totals_ms["speaker"] += speaker_ms

        self.state.timing_totals_ms["total"] += detect_ms + transcribe_ms + sentence_ms + speaker_ms
        self.state.window_ts = self.state.audio_ts
        self.state.acknowledged_ts = self.state.audio_ts
        self._emit_transcribing_progress(audio_stream)

        for i, utterance in enumerate(utterances):
            remaining = utterances[i + 1 :]
            self._emit_utterance_status(utterance=utterance, unacknowledged=remaining, unconfirmed=[])
            yield utterance, remaining, []

    def transcribe(
        self, audio_stream: AudioStream, working_prefix_no_ext: str = "out"
    ) -> Generator[Tuple[Utterance, List[Utterance], List[Word]], None, None]:
        self.state = State(self.config, working_prefix_no_ext=working_prefix_no_ext)
        try:
            if not self.config.code_switching:
                yield from self.transcribe_naive(audio_stream=audio_stream, working_prefix_no_ext=working_prefix_no_ext)
                return

            if audio_stream.start_offset != 0:
                self.state.window_ts = audio_stream.start_offset
                self.state.audio_ts = audio_stream.start_offset

            STATUS_LOG.info("Starting main loop for audio transcription.")
            self._emit_status(StatusUpdate(state="transcribing"))
            while True:
                has_more_audio = self.capture_audio(audio_source=audio_stream)
                had_utterances = False

                # capture any utterance that slipped out of the current window
                for utterance, unacknowledged, unconfirmed in self.flush_overflowing_utterances(diarization=audio_stream.diarization):
                    self._emit_utterance_status(
                        utterance=utterance,
                        unacknowledged=unacknowledged,
                        unconfirmed=unconfirmed,
                    )
                    yield utterance, unacknowledged, unconfirmed

                # attempt to acknowledge new utterances from current window
                for utterance, unacknowmedged, unconfirmed in self.process_audio_window(audio_stream=audio_stream):
                    had_utterances = True
                    self._emit_utterance_status(
                        utterance=utterance,
                        unacknowledged=unacknowmedged,
                        unconfirmed=unconfirmed,
                    )
                    yield utterance, unacknowmedged, unconfirmed

                self._emit_transcribing_progress(audio_stream)

                if not had_utterances and not has_more_audio:
                    break

        except KeyboardInterrupt:
            LOG.warning("KeyboardInterrupt detected, stopping transcription.")
            LOG.debug("Stopping...")
        # pylint: disable=broad-exception-caught
        except Exception as e:
            LOG.error(f"An unexpected error occurred: {e}\n{traceback.format_exc()}")
            LOG.debug("Stopping...")
        finally:
            LOG.info(
                (
                    "Pipeline totals: detect=%.1fms transcribe=%.1fms confirm=%.1fms sentence=%.1fms "
                    "acknowledge=%.1fms speaker=%.1fms output=%.1fms advance=%.1fms total=%.1fms"
                ),
                self.state.timing_totals_ms["detect"],
                self.state.timing_totals_ms["transcribe"],
                self.state.timing_totals_ms["confirm"],
                self.state.timing_totals_ms["sentence"],
                self.state.timing_totals_ms["acknowledge"],
                self.state.timing_totals_ms["speaker"],
                self.state.timing_totals_ms["output"],
                self.state.timing_totals_ms["advance"],
                self.state.timing_totals_ms["total"],
            )
            for i, utterance in enumerate(self.state.unacknowledged_utterances):
                utterance.speaker = self.assign_speaker(utterance, audio_stream.diarization)
                LOG.debug("Finalizing leftover pending utterance: %s", self._summarize_utterance(utterance))
                self._emit_utterance_status(
                    utterance=utterance,
                    unacknowledged=self.state.unacknowledged_utterances[i + 1 :],
                    unconfirmed=self.state.unconfirmed_words,
                )
                yield utterance, self.state.unacknowledged_utterances[i + 1 :], self.state.unconfirmed_words

            if len(self.state.unconfirmed_words) > 0:
                unconfirmed_utterance: Utterance = Utterance.from_words(
                    utterance_id=self.state.utterance_id.next(), words=self.state.unconfirmed_words
                )
                unconfirmed_utterance.speaker = self.assign_speaker(unconfirmed_utterance, audio_stream.diarization)
                LOG.debug("Finalizing leftover unconfirmed words: %s", self._summarize_utterance(unconfirmed_utterance))
                self._emit_utterance_status(
                    utterance=unconfirmed_utterance,
                    unacknowledged=[],
                    unconfirmed=[],
                )
                yield unconfirmed_utterance, [], []


def execute(
    *,
    config: Config,
    source_path: str,
    audio_sources: List[AudioSource],
    write_config: TranscriptWriterConfig,
    output_formats: List[str],
    output_prefix_no_ext: str,
    working_prefix_no_ext: str,
    eval_file: Optional[str],
    status_hook: Optional[StatusHook] = None,
):
    all_utterances: List[Utterance] = []
    transcriber = Verbatim(config, status_hook=status_hook)
    stdout_enabled = True
    for idx, audio_source in enumerate(audio_sources):
        STATUS_LOG.info("Transcribing from audio source: %s", audio_source.source_name)
        # When multiple sources are present (e.g., per-channel VTTM), isolate per-source outputs to avoid clobbering.
        per_source_prefix = output_prefix_no_ext
        if len(audio_sources) > 1:
            suffix = getattr(audio_source, "file_id", None) or f"src{idx}"
            per_source_prefix = f"{output_prefix_no_ext}-{suffix}"
        writer = configure_writers(
            write_config,
            output_formats=output_formats,
            original_audio_file=audio_source.source_name,
            output_prefix_no_ext=per_source_prefix,
            stdout_enabled=stdout_enabled,
        )
        writer.open()
        with audio_source.open() as audio_stream:
            for utterance, unacknowledged, unconfirmed in transcriber.transcribe(
                audio_stream=audio_stream, working_prefix_no_ext=working_prefix_no_ext
            ):
                writer.write(
                    utterance=utterance,
                    unacknowledged_utterance=unacknowledged,
                    unconfirmed_words=unconfirmed,
                )
                all_utterances.append(utterance)
        writer.close()
        STATUS_LOG.info("Done transcribing from audio source: %s", audio_source.source_name)

    if len(audio_sources) > 1:
        sorted_utterances: List[Utterance] = sorted(all_utterances, key=lambda x: x.start_ts)
        writer = configure_writers(
            write_config,
            output_formats=output_formats,
            original_audio_file=source_path,
            output_prefix_no_ext=output_prefix_no_ext,
            stdout_enabled=stdout_enabled,
        )
        writer.open()
        for sorted_utterance in sorted_utterances:
            writer.write(utterance=sorted_utterance)
        writer.close()

    if eval_file:
        LOG.info("Lazy-loading evaluation metrics.")
        from .eval.compare import compute_metrics  # pylint: disable=import-outside-toplevel

        sorted_utterances: List[Utterance] = sorted(all_utterances, key=lambda x: x.start_ts)
        ref_utterances: List[Utterance] = read_utterances(eval_file)
        metrics = compute_metrics(sorted_utterances, ref_utterances)
        print(metrics)



from dataclasses import dataclass
from typing import List, Literal, Optional, Protocol

from verbatim.transcript.words import Utterance, Word

StatusProgressUnit = Literal["percent", "count", "seconds"]


@dataclass(frozen=True)
class StatusProgress:
    current: float
    units: StatusProgressUnit
    start: float = 0.0
    finish: Optional[float] = None


@dataclass(frozen=True)
class StatusUpdate:
    state: Optional[str]
    progress: Optional[StatusProgress] = None
    utterance: Optional[Utterance] = None
    unacknowledged_utterances: Optional[List[Utterance]] = None
    unconfirmed_words: Optional[List[Word]] = None


class StatusHook(Protocol):
    def __call__(self, update: StatusUpdate) -> None: ...



import io
import logging
import os
from typing import Any, Dict, List, Optional

import numpy as np
import soundfile as sf

from verbatim.cache import ArtifactCache, FileBackedArtifactCache
from verbatim_audio.convert import convert_bytes_to_wav
from verbatim_diarization.diarize.base import DiarizationStrategy
from verbatim_diarization.utils import sanitize_uri_component
from verbatim_files.rttm import Annotation, Segment, dumps_rttm
from verbatim_files.vttm import AudioRef, dumps_vttm

LOG = logging.getLogger(__name__)


def _parse_bool(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        lowered = value.strip().lower()
        if lowered in ("1", "true", "yes", "y", "on"):
            return True
        if lowered in ("0", "false", "no", "n", "off"):
            return False
    return bool(value)


class SenkoDiarization(DiarizationStrategy):
    def __init__(
        self,
        *,
        cache: ArtifactCache,
        device: str,
        warmup: bool = True,
        quiet: bool = True,
        vad: str = "auto",
        clustering: str = "auto",
        accurate: Optional[bool | str] = None,
        mer_cos: Optional[float] = None,
    ):
        super().__init__(cache=cache)
        self.device = self._resolve_device(device)
        self.warmup = warmup
        self.quiet = quiet
        self.vad = vad
        self.clustering = clustering
        self.accurate = None if accurate is None else _parse_bool(accurate)
        self.mer_cos = mer_cos
        self._diarizer: Any = None

    @staticmethod
    def _resolve_device(device: str) -> str:
        if device == "mps":
            return "coreml"
        if device in ("cpu", "cuda"):
            return device
        return "auto"

    def _get_diarizer(self) -> Any:
        if self._diarizer is not None:
            return self._diarizer

        try:
            import senko  # pylint: disable=import-outside-toplevel
        except ImportError as exc:
            raise RuntimeError(
                "Senko diarization requires the Senko package. Install it with "
                "`uv pip install \"git+https://github.com/narcotic-sh/senko.git\"`."
            ) from exc

        self._diarizer = senko.Diarizer(
            device=self.device,
            vad=self.vad,
            clustering=self.clustering,
            warmup=self.warmup,
            quiet=self.quiet,
            mer_cos=self.mer_cos,
        )
        return self._diarizer

    @staticmethod
    def _is_compatible_wav(path: str) -> bool:
        if not path.lower().endswith(".wav") or not os.path.exists(path):
            return False

        info = sf.info(path)
        subtype = (info.subtype or "").upper()
        return info.samplerate == 16000 and info.channels == 1 and subtype == "PCM_16"

    def _prepare_wav_path(self, file_path: str) -> tuple[str, Optional[str]]:
        if self._is_compatible_wav(file_path):
            return file_path, None

        if not isinstance(self.cache, FileBackedArtifactCache) or not self.cache.base_dir:
            raise RuntimeError(
                "Senko diarization requires an on-disk 16kHz mono PCM WAV input. "
                "Use an already-normalized WAV file or provide --workdir so Verbatim can materialize a temporary WAV."
            )

        input_bytes = self.cache.get_bytes(file_path)
        if not input_bytes:
            raise RuntimeError(f"Cached bytes missing for input '{file_path}'. Populate the artifact cache before invoking Senko.")

        base_name = sanitize_uri_component(os.path.splitext(os.path.basename(file_path))[0], fallback="senko")
        temp_path = f"{base_name}-senko.wav"
        materialized_path = convert_bytes_to_wav(
            input_bytes=input_bytes,
            input_label=file_path,
            working_prefix_no_ext=os.path.join(self.cache.base_dir, f"{base_name}-senko"),
            preserve_channels=False,
            cache=self.cache,
        )
        return materialized_path or temp_path, materialized_path

    def _prepare_samples(self, file_path: str) -> np.ndarray:
        input_bytes = self.cache.get_bytes(file_path)
        if not input_bytes:
            raise RuntimeError(f"Cached bytes missing for input '{file_path}'. Populate the artifact cache before invoking Senko.")

        samples, sample_rate = sf.read(io.BytesIO(input_bytes), dtype="float32")
        if samples.ndim > 1:
            samples = samples.mean(axis=1)
        if sample_rate != 16000:
            raise RuntimeError(
                f"Senko in-memory diarization expects 16kHz audio from the artifact cache, got {sample_rate}Hz for '{file_path}'."
            )
        return samples.astype(np.float32, copy=False)

    @staticmethod
    def _segments_to_annotation(segments: List[Dict[str, Any]], *, file_id: str) -> Annotation:
        annotation_segments = []
        for segment in segments:
            start = float(segment["start"])
            end = float(segment["end"])
            speaker = str(segment["speaker"])
            annotation_segments.append(Segment(start=start, end=end, speaker=speaker, file_id=file_id))
        return Annotation(segments=annotation_segments, file_id=file_id)

    # pylint: disable=too-many-positional-arguments
    def compute_diarization(
        self,
        file_path: str,
        out_rttm_file: Optional[str] = None,
        out_vttm_file: Optional[str] = None,
        status_hook=None,
        nb_speakers: Optional[int] = None,
        **kwargs,
    ) -> Annotation:
        _ = status_hook
        if nb_speakers is not None:
            LOG.warning("Senko diarization does not accept a fixed speaker-count hint; ignoring nb_speakers=%s", nb_speakers)

        accurate = self.accurate
        if "accurate" in kwargs and kwargs["accurate"] is not None:
            accurate = _parse_bool(kwargs["accurate"])

        wav_path = None
        temp_path = None
        try:
            diarizer = self._get_diarizer()
            if self._is_compatible_wav(file_path):
                wav_path, temp_path = self._prepare_wav_path(file_path)
                result = diarizer.diarize(wav_path, accurate=accurate, generate_colors=False)
            elif callable(getattr(diarizer, "diarize_samples", None)) and not isinstance(self.cache, FileBackedArtifactCache):
                samples = self._prepare_samples(file_path)
                LOG.info("Using Senko in-memory diarization for %s", file_path)
                result = diarizer.diarize_samples(samples, sample_rate=16000, accurate=accurate, generate_colors=False)
            else:
                wav_path, temp_path = self._prepare_wav_path(file_path)
                result = diarizer.diarize(wav_path, accurate=accurate, generate_colors=False)
        except Exception as exc:
            raise RuntimeError(f"Senko diarization failed for '{file_path}': {exc}") from exc
        finally:
            if temp_path:
                self.cache.delete(temp_path)

        if result is None:
            return Annotation()

        merged_segments = result.get("merged_segments", [])
        file_id = sanitize_uri_component(os.path.splitext(os.path.basename(file_path))[0])
        annotation = self._segments_to_annotation(merged_segments, file_id=file_id)

        timing_stats = result.get("timing_stats")
        if timing_stats:
            LOG.info("Senko timing stats for %s: %s", file_path, timing_stats)

        if out_rttm_file:
            self.cache.set_text(out_rttm_file, dumps_rttm(annotation))

        if out_vttm_file:
            audio_refs = [AudioRef(id=file_id, path=file_path, channels=None)]
            self.cache.set_text(out_vttm_file, dumps_vttm(audio=audio_refs, annotation=annotation))

        return annotation



from dataclasses import dataclass
from typing import Optional

Annotation = object  # pylint: disable=invalid-name


@dataclass
class SourceConfig:
    isolate: Optional[bool] = None
    diarize_strategy: Optional[str] = None
    speakers: Optional[int] = None
    diarization: Optional[Annotation] = None  # verbatim_files.rttm.Annotation expected at runtime
    diarization_file: Optional[str] = None  # legacy RTTM; when set, we derive from vttm
    vttm_file: Optional[str] = None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Optional

from verbatim_audio.audio import samples_to_seconds

if TYPE_CHECKING:
    from faster_whisper.transcribe import Word as WhisperWord


@dataclass
class Word:
    start_ts: int
    end_ts: int
    word: str
    probability: float
    lang: str

    @classmethod
    def from_word(cls, word: "WhisperWord", lang: str, ts_offset: int = 0) -> "Word":
        """Creates a Word instance from a Word object with a timestamp offset."""
        # Lazy import to avoid pulling faster_whisper during CLI startup.

        start_ts = int(word.start * 16000) + ts_offset
        end_ts = int(word.end * 16000) + ts_offset
        return cls(
            start_ts=start_ts,
            lang=lang,
            end_ts=end_ts,
            word=word.word,
            probability=word.probability,
        )


@dataclass
class Utterance:
    utterance_id: str
    speaker: Optional[str]
    start_ts: int
    end_ts: int
    text: str
    words: List[Word] = field(default_factory=list)

    def get_start(self) -> float:
        return samples_to_seconds(self.start_ts)

    def get_end(self) -> float:
        return samples_to_seconds(self.end_ts)

    @classmethod
    def from_words(cls, utterance_id: str, words: List[Word], speaker: Optional[str] = None) -> "Utterance":
        start_ts = words[0].start_ts
        end_ts = words[-1].end_ts
        text = "".join([w.word for w in words])
        return cls(utterance_id=utterance_id, start_ts=start_ts, end_ts=end_ts, words=words, text=text, speaker=speaker)

    @classmethod
    def marker(cls, utterance_id: str, start_ts: int, end_ts: int, text: str, *, speaker: Optional[str] = None) -> "Utterance":
        return cls(utterance_id=utterance_id, speaker=speaker, start_ts=start_ts, end_ts=end_ts, text=text, words=[])



import logging
from abc import ABC, abstractmethod
from typing import Optional

from verbatim.cache import ArtifactCache
from verbatim.status_types import StatusHook
from verbatim_files.rttm import Annotation, dumps_rttm

LOG = logging.getLogger(__name__)


class DiarizationStrategy(ABC):
    """Base class for all diarization strategies"""

    def __init__(self, *, cache: ArtifactCache):
        self.cache = cache

    @abstractmethod
    def compute_diarization(
        self,
        file_path: str,
        out_rttm_file: Optional[str] = None,
        out_vttm_file: Optional[str] = None,
        status_hook: Optional[StatusHook] = None,
        **kwargs,
    ) -> Annotation:
        """
        Compute speaker diarization for the given audio file.

        Args:
            file_path: Path to the audio file
            out_rttm_file: Optional legacy RTTM output path (derived from VTTM if present)
            out_vttm_file: Preferred VTTM output path
            **kwargs: Strategy-specific parameters

        Returns:
            Annotation object containing speaker segments
        """

    def save_rttm(self, annotation: Annotation, out_rttm_file: str):
        """Save annotation to RTTM file"""
        self.cache.set_text(out_rttm_file, dumps_rttm(annotation))



from __future__ import annotations

from typing import TYPE_CHECKING, Union

from verbatim_files.rttm import Annotation, Segment

if TYPE_CHECKING:
    from pyannote.core.annotation import Annotation as PyannoteAnnotation
else:  # pragma: no cover - typing-only alias when pyannote is optional
    PyannoteAnnotation = object  # type: ignore[misc]


def to_rttm_annotation(annotation: Union[Annotation, "PyannoteAnnotation"]) -> Annotation:
    if isinstance(annotation, Annotation):
        return annotation
    if hasattr(annotation, "segments"):
        try:
            return Annotation(segments=list(annotation.segments), file_id=getattr(annotation, "file_id", None))
        except Exception as exc:  # pragma: no cover - defensive: unexpected segments shape
            raise TypeError("Unsupported diarization annotation segments") from exc
    if hasattr(annotation, "itertracks"):
        uri = getattr(annotation, "uri", None)
        file_id = str(uri) if uri else ""
        segments = [
            Segment(start=seg.start, end=seg.end, speaker=str(label), file_id=file_id)
            for seg, _track, label in annotation.itertracks(yield_label=True)  # type: ignore[attr-defined]
        ]
        return Annotation(segments=segments, file_id=file_id or None)
    raise TypeError("Unsupported diarization annotation type for RTTM serialization")



import logging
import math
import os
from typing import BinaryIO, Optional

# pylint: disable=c-extension-no-member
import av
import numpy as np
from numpy.typing import NDArray

from ..audio import seconds_to_samples
from .audiosource import AudioSource, AudioStream

LOG = logging.getLogger(__name__)


class PyAVAudioStream(AudioStream):
    source: "PyAVAudioSource"

    def __init__(self, source: "PyAVAudioSource"):
        """Open the container, find the audio stream, and seek if needed."""
        super().__init__(start_offset=seconds_to_samples(source.start_time), diarization=None)
        self.source = source

        # Internals
        # pylint: disable=c-extension-no-member
        self._container: av.container.InputContainer = None  # pyright: ignore[reportAttributeAccessIssue]
        self._stream: av.audio.stream.AudioStream = None  # pyright: ignore[reportAttributeAccessIssue]

        # Buffer for leftover samples (when frames don't line up exactly with chunk size)
        self._sample_buffer = np.array([], dtype=np.float32)

        self._closed = False
        self._done_decoding = False

        if self.source.file_obj is not None:
            LOG.debug("Opening cached bytes with PyAV: %s", self.source.file_path)
            self._container = av.open(self.source.file_obj, format=self.source.format_hint)
        else:
            LOG.debug("Opening file with PyAV: %s", self.source.file_path)
            self._container = av.open(self.source.file_path)

        # Find the first audio stream (or choose a specific one if needed)
        audio_streams = [s for s in self._container.streams if s.type == "audio"]
        if not audio_streams:
            raise ValueError("No audio streams found in file.")

        self._stream = audio_streams[0]
        LOG.debug("Audio stream channels: %s", self._stream.channels)
        LOG.debug("Preserve channels: %s", self.source.preserve_channels)
        self._stream.thread_type = "AUTO"  # allow FFmpeg to use threading if beneficial

        # If you want to force a certain sample format, channel layout, etc.,
        # you can set those on self._stream or handle them in code below.
        # E.g., self._stream.codec_context.sample_rate = self.target_sample_rate

        # If we want to seek to `start_time`, do so in seconds:
        if self.source.start_time > 0:
            # av.seek() or container.seek() uses timestamps in AV_TIME_BASE units
            # but PyAV usually accepts "seconds" if we specify 'any' for backward flag
            self._container.seek(
                int(self.source.start_time / self._stream.time_base),
                any_frame=False,
                stream=self._stream,
            )
            LOG.debug("Seeking to %s seconds.", self.source.start_time)

        # We create a generator that decodes frames from the audio stream
        # This is the raw frames from the container
        frame_iter = self._container.decode(self._stream)
        if frame_iter is None:
            raise RuntimeError("Frame iterator is not initialized. Decoding cannot proceed.")
        self._frame_iter = frame_iter

        self._done_decoding = False
        self._closed = False

    def next_chunk(self, chunk_length=1) -> NDArray:
        """
        Return `chunk_length` seconds of audio as a NumPy array.
        This version makes the resampling and reshaping behavior predictable and explicit.
        """

        if self._closed:
            LOG.warning("next_chunk() called after close(). Returning empty array.")
            return np.array([], dtype=np.float32)

        if self._done_decoding and len(self._sample_buffer) == 0:
            return np.array([], dtype=np.float32)

        # --- Step 1: Define original and target audio parameters ---
        original_sample_rate = self._stream.codec_context.sample_rate
        original_channels = self._stream.codec_context.channels

        target_sample_rate = self.source.target_sample_rate
        target_channels = original_channels if self.source.preserve_channels else 1

        LOG.debug(f"Original sample rate: {original_sample_rate} Hz, channels: {original_channels}")
        LOG.debug(f"Target sample rate: {target_sample_rate} Hz, channels: {target_channels}")

        # --- Step 2: Determine how many samples are needed for this chunk ---
        needed_target_samples = int(chunk_length * target_sample_rate)
        needed_input_samples = math.ceil(needed_target_samples * original_sample_rate / target_sample_rate)
        LOG.debug(f"Want {needed_target_samples} output samples → need ~{needed_input_samples} input samples")

        # --- Step 3: Decode and collect raw frames as-is ---
        collected = []

        total_collected_samples = 0
        while total_collected_samples < needed_input_samples and not self._done_decoding:
            try:
                frame = next(self._frame_iter)
            except StopIteration:
                LOG.debug("Reached end of stream")
                self._done_decoding = True
                break

            # Respect end_time if set
            if self.source.end_time is not None and frame.pts is not None:
                timestamp_sec = float(frame.pts * self._stream.time_base)
                if timestamp_sec > self.source.end_time:
                    LOG.debug("End time %ss reached (current=%.2fs)", self.source.end_time, timestamp_sec)
                    self._done_decoding = True
                    break

            # Convert the frame to ndarray, float32
            arr = frame.to_ndarray().astype(np.float32, copy=False)

            # Normalize shape to always be (channels, samples)
            if arr.ndim == 1:
                arr = arr[np.newaxis, :]  # mono → (1, samples)
            elif arr.ndim == 2:
                pass  # already (channels, samples)
            else:
                raise ValueError(f"Unexpected frame shape: {arr.shape}")

            total_collected_samples += arr.shape[1]
            collected.append(arr)

        # --- Step 4: Stack collected frames into one array ---
        if not collected:
            return np.array([], dtype=np.float32)

        full_array = np.concatenate(collected, axis=1)  # shape: (channels, total_samples)
        LOG.debug(f"Concatenated shape before mixing: {full_array.shape}")

        # --- Optional: Downmix to mono if needed ---
        if target_channels == 1 and full_array.shape[0] > 1:
            # Average across channels: axis 0 is channels
            full_array = full_array.mean(axis=0, keepdims=True)  # shape becomes (1, total_samples)
            LOG.debug(f"Downmixed to mono: {full_array.shape}")

        # --- Step 5: Resample if needed ---
        if original_sample_rate != target_sample_rate:
            LOG.debug(f"Resampling from {original_sample_rate} → {target_sample_rate} Hz")

            # Resample each channel independently
            up = int(target_sample_rate)
            down = int(original_sample_rate)

            # Lazy import to avoid pulling scipy.signal during CLI startup
            from scipy.signal import resample_poly  # type: ignore  # pylint: disable=import-outside-toplevel

            resampled_channels = []
            for ch_idx in range(full_array.shape[0]):
                resampled = resample_poly(full_array[ch_idx], up, down)
                resampled_channels.append(resampled)

            # Stack back into (channels, samples)
            full_array = np.stack(resampled_channels, axis=0)
            LOG.debug(f"Resampled shape: {full_array.shape}")
            LOG.debug(f"Chunk duration: {chunk_length} s → {needed_target_samples} samples needed at target rate")

        if target_channels == 1:
            return np.squeeze(full_array)
        else:
            return full_array.T

    def has_more(self) -> bool:
        """
        Return True if there's more data to read from the stream,
        or if there's still leftover samples in the buffer.
        """
        if self._closed:
            return False
        if not self._done_decoding:
            return True
        # If done decoding, but still have samples in buffer, it's True
        return len(self._sample_buffer) > 0

    def close(self):
        """Close the container and release resources."""
        if self._container and not self._closed:
            LOG.debug("Closing PyAV container.")
            self._container.close()
        self._closed = True
        self._done_decoding = True
        self._sample_buffer = np.array([], dtype=np.float32)

    def get_nchannels(self) -> int:
        return self._stream.channels

    def get_rate(self) -> int:
        return self.source.target_sample_rate


class PyAVAudioSource(AudioSource):
    """
    A chunk-based audio reader that uses PyAV to decode audio frames.

    - Reads from a container (any format that FFmpeg/PyAV supports).
    - Decodes audio frames into an internal buffer.
    - Returns them in "chunk_length" second increments via next_chunk().
    - By default, it streams as 16-bit int PCM at some target rate (e.g. 16 kHz).
    """

    def __init__(
        self,
        *,
        file_path: str,
        file_obj: Optional[BinaryIO] = None,
        target_sample_rate: int = 16000,
        start_time: float = 0.0,
        end_time: Optional[float] = None,
        preserve_channels: bool = False,
    ):
        """
        :param file_path: Path/URL to audio file
        :param file_obj: File-like object containing encoded audio bytes
        :param target_sample_rate: Desired sample rate for output (e.g. 16k)
        :param start_time: Seek to this time (seconds) before reading
        :param end_time: Stop reading after this time (seconds) from start
        :param preserve_channels: If True, return audio as-is (no channel mixing)
        """
        super().__init__(source_name=file_path)
        self.file_path = file_path
        self.file_obj = file_obj
        self.target_sample_rate = target_sample_rate
        self.start_time = start_time
        self.end_time = end_time
        self.preserve_channels = preserve_channels
        self.format_hint = None
        if self.file_obj is not None:
            ext = os.path.splitext(self.file_path)[-1].lstrip(".").lower()
            if ext:
                format_map = {
                    "m4a": "mp4",
                    "m4b": "mp4",
                    "m4p": "mp4",
                    "m4v": "mp4",
                }
                self.format_hint = format_map.get(ext, ext)

    def open(self):
        return PyAVAudioStream(source=self)



import io
import os

from verbatim.cache import ArtifactCache


def convert_to_wav(
    input_path: str,
    working_prefix_no_ext: str,
    preserve_channels: bool = False,
    overwrite=True,
    *,
    cache: ArtifactCache,
) -> str:
    # pylint: disable=import-outside-toplevel
    from .sources.ffmpegfileaudiosource import PyAVAudioSource
    from .sources.wavsink import WavSink

    converted_path = working_prefix_no_ext + ".wav"

    if not overwrite and os.path.exists(converted_path) is True:
        return converted_path

    temp_file_audio_source = PyAVAudioSource(file_path=input_path, preserve_channels=preserve_channels)
    WavSink.dump_to_wav(
        audio_source=temp_file_audio_source,
        output_path=converted_path,
        preserve_channels=preserve_channels,
        cache=cache,
    )

    return converted_path


def convert_bytes_to_wav(
    *,
    input_bytes: bytes,
    input_label: str,
    working_prefix_no_ext: str,
    preserve_channels: bool = False,
    overwrite=True,
    cache: ArtifactCache,
) -> str:
    # pylint: disable=import-outside-toplevel
    from .sources.ffmpegfileaudiosource import PyAVAudioSource
    from .sources.wavsink import WavSink

    converted_path = working_prefix_no_ext + ".wav"

    if not overwrite and os.path.exists(converted_path) is True:
        return converted_path

    temp_file_audio_source = PyAVAudioSource(
        file_path=input_label,
        file_obj=io.BytesIO(input_bytes),
        preserve_channels=preserve_channels,
    )
    WavSink.dump_to_wav(
        audio_source=temp_file_audio_source,
        output_path=converted_path,
        preserve_channels=preserve_channels,
        cache=cache,
    )

    return converted_path



# pylint: disable=wrong-import-position
import os

# Disable pyannote telemetry unless caller opts in explicitly via env.
os.environ.setdefault("PYANNOTE_METRICS_ENABLED", "false")

from .diarize import PyAnnoteDiarization, PyAnnoteSeparationDiarization  # noqa: E402
from .separate import PyannoteSpeakerSeparation  # noqa: E402

__all__ = ["PyAnnoteDiarization", "PyAnnoteSeparationDiarization", "PyannoteSpeakerSeparation"]



import dataclasses
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional

from verbatim.transcript.words import Utterance, Word


# pylint: disable=invalid-name
class StyleEnum(Enum):
    """Enum that stringifies to its name for nicer CLI output."""

    def __str__(self) -> str:
        return self.name


class TimestampStyle(StyleEnum):
    none = 1
    start = 2
    range = 3
    minute = 4


class SpeakerStyle(StyleEnum):
    none = 1
    change = 2
    always = 3


class ProbabilityStyle(StyleEnum):
    none = 1
    line = 2
    line_75 = 3
    line_50 = 4
    line_25 = 5
    word = 6
    word_75 = 7
    word_50 = 8
    word_25 = 9


class LanguageStyle(StyleEnum):
    none = 1
    change = 2
    always = 3


@dataclasses.dataclass
class TranscriptWriterConfig:
    timestamp_style: TimestampStyle = TimestampStyle.none
    speaker_style: SpeakerStyle = SpeakerStyle.none
    probability_style: ProbabilityStyle = ProbabilityStyle.none
    language_style: LanguageStyle = LanguageStyle.none
    verbose: bool = False


class TranscriptWriter(ABC):
    def __init__(self, config: TranscriptWriterConfig):
        self.config = config

    def format_start(self) -> bytes:
        return b""

    @abstractmethod
    def format_utterance(
        self,
        utterance: Utterance,
        unacknowledged_utterance: Optional[List[Utterance]] = None,
        unconfirmed_words: Optional[List[Word]] = None,
    ) -> bytes:
        pass

    def format_end(self) -> bytes:
        return b""

    def flush(self) -> bytes:
        return b""

    def post_close(self, _output_path: str) -> None:
        return None



import logging
import os
from typing import Optional, Tuple

import librosa
import numpy as np
from audio_separator.separator import Separator
from numpy.typing import NDArray
from scipy.io.wavfile import write as wav_write

from verbatim_audio.audio import format_audio

# Configure logger
LOG = logging.getLogger(__name__)


class VoiceIsolation:
    def __init__(self, log_level: int = logging.WARN, model_name: str = "MDX23C-8KFFT-InstVoc_HQ_2.ckpt"):
        self.separator = Separator(log_level=log_level, sample_rate=16000)
        cache_root = os.getenv("VERBATIM_MODEL_CACHE")
        offline_env = os.getenv("VERBATIM_OFFLINE", "0").lower() in ("1", "true", "yes")

        # If a cache is configured, prefer a cached checkpoint path under it
        model_path = model_name
        if cache_root and not os.path.isabs(model_name):
            candidate = os.path.join(cache_root, "audio-separator", model_name)
            if os.path.exists(candidate):
                model_path = candidate
            elif offline_env:
                raise RuntimeError(f"Offline mode is enabled and isolation model '{model_name}' was not found at {candidate}")

        self.separator.load_model(model_path)

    def __enter__(self) -> "VoiceIsolation":
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        del self.separator
        return False

    def isolate_voice_in_file(self, file: str, out_voice: Optional[str] = None, out_noise: Optional[str] = None) -> Tuple[str, str]:
        if out_voice is None:
            out_voice = "debug_audio-vocals"
        if out_noise is None:
            out_noise = "debug_audio-instrumental"

        # Use MDX to separate vocals from the source audio
        output_file_paths = self.separator.model_instance.separate(  # pyright: ignore[reportOptionalMemberAccess]
            file,
            {"Instrumental": out_noise, "Vocals": out_voice},
        )

        # The separated files: instrument track and vocal track
        instrument_audio_path = output_file_paths[0]
        voice_audio_path = output_file_paths[1]

        return voice_audio_path, instrument_audio_path

    def isolate_voice_in_array(self, audio: NDArray) -> NDArray:
        input_length = len(audio)

        if not np.any(audio):
            # audio is empty, skip
            return audio

        # Save the input audio to a temporary file
        temp_audio_file = "voice-isolation.wav"
        wav_write(temp_audio_file, 16000, audio)

        voice_audio_path, _ = self.isolate_voice_in_file(file=temp_audio_file)

        # Load the vocal audio back into a NumPy array
        voice_audio, voice_sampling_rate = librosa.load(voice_audio_path, sr=None, mono=False)
        voice_audio = voice_audio.T  # librosa formats (nchannel, samples) and we expect (samples, nchannels)

        # Format the vocal audio to mono and 16 kHz
        formatted_voice_audio = format_audio(voice_audio, int(voice_sampling_rate))
        wav_write("voice-isolation-filtered.wav", 16000, formatted_voice_audio)

        # ensure output has same length as input
        if len(formatted_voice_audio) < input_length:
            padding_length = input_length - len(formatted_voice_audio)
            formatted_voice_audio = np.pad(
                array=formatted_voice_audio,
                pad_width=(0, padding_length),
                mode="constant",
                constant_values=0,
            )
        elif len(formatted_voice_audio) > input_length:
            formatted_voice_audio = formatted_voice_audio[:input_length]

        return formatted_voice_audio



def format_milliseconds(ms: float) -> str:
    # Calculate hours, minutes, and seconds
    seconds_total = int(ms // 1000)
    hours = int(seconds_total // 3600)
    minutes = int((seconds_total % 3600) // 60)
    seconds = int(seconds_total) % 60

    return f"{hours:02}:{minutes:02}:{seconds:02}"



import io
import wave

import numpy as np

from verbatim.cache import ArtifactCache

from .audiosource import AudioSource


class WavSink:
    @staticmethod
    def dump_to_wav(
        audio_source: AudioSource,
        output_path: str,
        *,
        cache: ArtifactCache,
        sample_rate: int = 16000,
        preserve_channels: bool = False,
    ):
        with audio_source.open() as audio_stream:
            input_sample_rate = audio_stream.get_rate()  # e.g., 48000
            num_channels = audio_stream.get_nchannels() if preserve_channels else 1
            sample_width = 2  # 16-bit PCM

            buffer = io.BytesIO()
            with wave.open(buffer, "w") as wav_file:
                # pylint: disable=no-member
                wav_file.setnchannels(num_channels)
                wav_file.setsampwidth(sample_width)
                wav_file.setframerate(sample_rate)

                while audio_stream.has_more():
                    audio_chunk: np.ndarray = np.asarray(audio_stream.next_chunk(chunk_length=1), dtype=np.float32)

                    # If stereo and not preserving channels, downmix to mono
                    if not preserve_channels and audio_chunk.ndim > 1:
                        audio_chunk = audio_chunk.mean(axis=1)

                    # Resample the audio if needed
                    if input_sample_rate != sample_rate:
                        # Lazy import to avoid pulling scipy.signal during CLI startup
                        from scipy.signal import resample  # type: ignore  # pylint: disable=import-outside-toplevel

                        target_len = int(len(audio_chunk) * sample_rate / input_sample_rate)
                        audio_chunk = np.asarray(resample(audio_chunk, target_len))

                    # Convert to 16-bit PCM
                    int_samples = (audio_chunk * 32767).clip(-32768, 32767).astype(np.int16)
                    wav_file.writeframes(int_samples.tobytes())
            cache.set_bytes(output_path, buffer.getvalue())



import logging
from time import perf_counter

from numpy.typing import NDArray

from .interfaces import GuessLanguageFn, LanguageDetectionRequest, LanguageDetectionResult

LOG = logging.getLogger(__name__)


def _slice_audio(audio: NDArray, start: int, duration: int) -> NDArray:
    return audio[start : start + duration]


def detect_language(request: LanguageDetectionRequest, guess_fn: GuessLanguageFn) -> LanguageDetectionResult:
    """Stateless language detection helper that can be shared by verbatim-core."""
    if len(request.lang) == 0:
        LOG.warning("Language is not set - defaulting to english")
        return LanguageDetectionResult(language="en", probability=1.0, samples_used=0)

    if len(request.lang) == 1:
        return LanguageDetectionResult(language=request.lang[0], probability=1.0, samples_used=0)

    lang_sample_start = max(0, request.timestamp - request.window_ts)
    available_samples = request.audio_ts - request.window_ts - lang_sample_start
    lang_samples_size = min(2 * 16000, available_samples)
    detect_start = perf_counter()

    while True:
        lang_samples = _slice_audio(request.audio, lang_sample_start, lang_samples_size)
        lang, prob = guess_fn(lang_samples, request.lang)
        if prob > 0.5 or lang_samples_size == available_samples:
            break
        # retry with larger sample
        lang_samples_size = min(2 * lang_samples_size, available_samples)

    elapsed_ms = (perf_counter() - detect_start) * 1000.0
    LOG.info(
        "Detected language using %s samples (%.3fs): lang=%s prob=%.3f elapsed=%.1fms",
        lang_samples_size,
        lang_samples_size / 16000.0,
        lang,
        prob,
        elapsed_ms,
    )

    return LanguageDetectionResult(language=lang, probability=prob, samples_used=lang_samples_size)



import logging
import queue

import numpy as np
import pyaudio
import sounddevice as sd
from numpy.typing import NDArray

from ..audio import samples_to_seconds
from .audiosource import AudioSource, AudioStream

LOG = logging.getLogger(__name__)


class MicAudioStreamSoundDevice(AudioStream):
    source: "MicAudioSourceSoundDevice"
    audio_queue: queue.Queue
    stream: sd.InputStream

    def __init__(self, source: "MicAudioSourceSoundDevice"):
        super().__init__(start_offset=0, diarization=None)
        self.source = source
        self.audio_queue = queue.Queue()

        # Open the audio stream with the callback
        self.stream = sd.InputStream(
            samplerate=self.source.sampling_rate,
            channels=1,
            blocksize=self.source.frames_per_buffer,
            callback=self._audio_callback,
        )
        self.stream.start()
        LOG.info("Audio stream opened.")

    # pylint: disable=unused-argument
    def _audio_callback(self, indata, frames, time, status: sd.CallbackFlags):
        if status:
            LOG.warning(f"Audio stream status: {status}")
            if status.input_overflow:
                LOG.error("Input overflow occurred! Some audio data was lost.")
            if status.input_underflow:
                LOG.error("Input underflow occurred!")  # Add the audio data to the queue
        chunk = indata.copy().ravel()
        self.audio_queue.put(chunk)
        # LOG.debug(f"Captured new audio: len={len(chunk)} min={min(chunk)} max={max(chunk)}")

    def close(self):
        if self.stream:
            self.stream.stop()
            self.stream.close()
            LOG.info("Audio stream closed.")

    def next_chunk(self, chunk_length=1) -> NDArray:
        """Fetch all available audio data from the queue without blocking."""
        frames = []
        while not self.audio_queue.empty():
            frames.append(self.audio_queue.get())

        if frames:
            # Concatenate all collected frames into a single numpy array
            audio_array = np.concatenate(frames, axis=0)
            # Convert the audio data to float32 and normalize to [-1, 1]
            audio_array = audio_array.astype(np.float32)
            LOG.debug(f"Fetched {len(audio_array)} ({samples_to_seconds(len(audio_array))}) samples.")
            return audio_array

        else:
            # Return an empty array if no data is available
            return np.array([], dtype=np.float32)

    def has_more(self):
        return True

    def get_nchannels(self) -> int:
        stream: sd.InputStream = self.stream
        idevice, _odevice = stream.channels
        return idevice

    def get_rate(self) -> int:
        # pylint: disable=protected-access
        return self.stream._samplerate


class MicAudioSourceSoundDevice(AudioSource):
    def __init__(self, sampling_rate: int = 16000, frames_per_buffer: int = 1024):
        super().__init__(source_name="<mic>")
        self.sampling_rate = sampling_rate
        self.frames_per_buffer = frames_per_buffer

    def open(self):
        return MicAudioStreamSoundDevice(source=self)


class MicAudioStreamPyAudio(AudioStream):
    source: "MicAudioSourcePyAudio"
    p: pyaudio.PyAudio
    stream: pyaudio.Stream
    nchannels: int = 1

    def __init__(self, source: "MicAudioSourcePyAudio", nchannels: int = 1):
        super().__init__(start_offset=0, diarization=None)
        self.source = source
        self.p: pyaudio.PyAudio = pyaudio.PyAudio()
        self.stream: pyaudio.Stream = self.p.open(
            format=pyaudio.paInt16,
            channels=nchannels,
            rate=self.source.sampling_rate,
            input=True,
            frames_per_buffer=self.source.frames_per_buffer,
        )

    def next_chunk(self, chunk_length=1) -> NDArray:
        LOG.info(f"Recording {chunk_length} seconds of audio.")
        frames = []
        # Read exactly chunk_length seconds of audio
        for _ in range(
            0,
            int(self.source.frames_per_iter / self.source.frames_per_buffer * chunk_length),
        ):
            data = self.stream.read(self.source.frames_per_buffer)
            frames.append(data)

        audio_bytes = b"".join(frames)
        audio_array = np.frombuffer(audio_bytes, dtype=np.int16)
        # Convert int16 array to float32 and normalize to [-1, 1]
        audio_array = audio_array.astype(np.float32) / 32768.0
        LOG.info("Finished recording audio chunk.")
        return audio_array

    def close(self):
        self.stream.stop_stream()
        self.stream.close()
        self.p.terminate()

    def has_more(self):
        return True

    def get_nchannels(self) -> int:
        return self.nchannels

    def get_rate(self) -> int:
        return self.source.sampling_rate


class MicAudioSourcePyAudio(AudioSource):
    frames_per_iter: int
    frames_per_buffer: int
    sampling_rate: int

    def __init__(
        self,
        latency: int = 16000,
        frames_per_buffer: int = 1000,
        sampling_rate: int = 16000,
    ):
        super().__init__(source_name="<mic>")
        self.frames_per_iter: int = latency
        self.frames_per_buffer: int = frames_per_buffer
        self.sampling_rate = sampling_rate

    def open(self) -> MicAudioStreamPyAudio:
        return MicAudioStreamPyAudio(source=self)
